# -*- coding: utf-8 -*-

# Copyright 1996-2015 PSERC. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

# DEBUG = False
# DEBUG = True

from numba import jit
# import numba as nb

# numpy is only used for debugging
# if DEBUG:
#     from numpy import zeros


# @jit(i8(c16[:], c16[:], i4[:], i4[:], i8[:], i8[:], f8[:], i8[:], i8[:]), nopython=True, cache=False)
@jit(nopython=True, cache=False)
def create_J(dVm_x, dVa_x, Yp, Yj, pvpq_lookup, refpvpq, pvpq, pq, Jx, Jj, Jp, slack_weights):  # pragma: no cover
    """Calculates Jacobian faster with numba and sparse matrices.

        Input: dS_dVa and dS_dVm in CSR sparse form (Yx = data, Yp = indptr, Yj = indices), pvpq, pq from pypower

        OUTPUT:  data from CSR form of Jacobian (Jx, Jj, Jp) and number of non zeros (nnz)

        @author: Florian Schaefer

        Calculate Jacobian entries

        J11 = dS_dVa[array([pvpq]).T, pvpq].real
        J12 = dS_dVm[array([pvpq]).T, pq].real
        J21 = dS_dVa[array([pq]).T, pvpq].imag
        J22 = dS_dVm[array([pq]).T, pq].imag

        Explanation of code:
        To understand the concept the CSR storage method should be known. See:
        https://de.wikipedia.org/wiki/Compressed_Row_Storage

        J has the shape
        | J11 | J12 |               | (pvpq, pvpq) | (pvpq, pq) |
        | --------- | = dimensions: | ------------------------- |
        | J21 | J22 |               |  (pq, pvpq)  |  (pq, pq)  |

        We first iterate the rows of J11 and J12 (for r in range lpvpq) and add the entries which are stored in dS_dV
        Then we iterate the rows of J21 and J22 (for r in range lpq) and add the entries from dS_dV

        Note: The row and column pointer of of dVm and dVa are the same as the one from Ybus
    """
    # Jacobi Matrix in sparse form
    # Jp, Jx, Jj equal J like:
    # J = zeros(shape=(ndim, ndim), dtype=float64)

    # get length of vectors
    lpvpq = len(pvpq)
    lpq = len(pq)
    lpv = lpvpq - lpq

    # nonzeros in J
    nnz = 0

    # iterate rows of J
    # first iterate pvpq (J11 and J12)
    for r in range(lpvpq):
        # nnzStar is necessary to calculate nonzeros per row
        nnzStart = nnz
        # iterate columns of J11 = dS_dVa.real at positions in pvpq
        # check entries in row pvpq[r] of dS_dV
        for c in range(Yp[pvpq[r]], Yp[pvpq[r] + 1]):
            # check if column Yj is in pvpq
            cc = pvpq_lookup[Yj[c]]
            # entries for J11 and J12
            if pvpq[cc] == Yj[c]:
                # entry found
                # equals entry of J11: J[r,cc] = dVa_x[c].real
                Jx[nnz] = dVa_x[c].real
                Jj[nnz] = cc
                nnz += 1
                # if entry is found in the "pq part" of pvpq = add entry of J12
                if cc >= lpv:
                    Jx[nnz] = dVm_x[c].real
                    Jj[nnz] = cc + lpq
                    nnz += 1
        # Jp: number of nonzeros per row = nnz - nnzStart (nnz at begging of loop - nnz at end of loop)
        Jp[r + 1] = nnz - nnzStart + Jp[r]
    # second: iterate pq (J21 and J22)
    for r in range(lpq):
        nnzStart = nnz
        # iterate columns of J21 = dS_dVa.imag at positions in pvpq
        for c in range(Yp[pq[r]], Yp[pq[r] + 1]):
            cc = pvpq_lookup[Yj[c]]
            if pvpq[cc] == Yj[c]:
                # entry found
                # equals entry of J21: J[r + lpvpq, cc] = dVa_x[c].imag
                Jx[nnz] = dVa_x[c].imag
                Jj[nnz] = cc
                nnz += 1
                if cc >= lpv:
                    # if entry is found in the "pq part" of pvpq = Add entry of J22
                    Jx[nnz] = dVm_x[c].imag
                    Jj[nnz] = cc + lpq
                    nnz += 1
        # Jp: number of nonzeros per row = nnz - nnzStart (nnz at begging of loop - nnz at end of loop)
        Jp[r + lpvpq + 1] = nnz - nnzStart + Jp[r + lpvpq]


# @jit(i8(c16[:], c16[:], i4[:], i4[:], i8[:], i8[:], f8[:], i8[:], i8[:]), nopython=True, cache=True)
@jit(nopython=True, cache=False)
def create_J2(dVm_x, dVa_x, Yp, Yj, pvpq_lookup, refpvpq, pvpq, pq, Jx, Jj, Jp, slack_weights):  # pragma: no cover
    """Calculates Jacobian faster with numba and sparse matrices. This version is similar to create_J except that
        if pvpq = pq (when no pv bus is available) some if statements are obsolete and J11 = J12 and J21 = J22

        Input: dS_dVa and dS_dVm in CSR sparse form (Yx = data, Yp = indptr, Yj = indices), pvpq, pq from pypower

        OUTPUT: data from CSR form of Jacobian (Jx, Jj, Jp) and number of non zeros (nnz)

        @author: Florian Schaefer
        @date: 30.08.2016

        see comments in create_J
    """
    # Jacobi Matrix in sparse form
    # Jp, Jx, Jj equal J like:
    # J = zeros(shape=(ndim, ndim), dtype=float64)

    # get info of vector
    lpvpq = len(pvpq)

    # nonzeros in J
    nnz = 0

    # iterate rows of J
    # first iterate pvpq (J11 and J12)
    for r in range(lpvpq):
        # nnzStart is necessary to calculate nonzeros per row
        nnzStart = nnz
        # iterate columns of J11 = dS_dVa.real at positions in pvpq
        # iterate columns of J12 = dS_dVm.real at positions in pq (=pvpq)
        for c in range(Yp[pvpq[r]], Yp[pvpq[r] + 1]):
            cc = pvpq_lookup[Yj[c]]
            if pvpq[cc] == Yj[c]:
                # entry found J11
                # J[r,cc] = dVa_x[c].real
                Jx[nnz] = dVa_x[c].real
                Jj[nnz] = cc
                nnz += 1
                # also entry in J12
                Jx[nnz] = dVm_x[c].real
                Jj[nnz] = cc + lpvpq
                nnz += 1
        # Jp: number of nonzeros per row = nnz - nnzStart (nnz at begging of loop - nnz at end of loop)
        Jp[r + 1] = nnz - nnzStart + Jp[r]
    # second: iterate pq (J21 and J22)
    for r in range(lpvpq):
        nnzStart = nnz
        # iterate columns of J21 = dS_dVa.imag at positions in pvpq
        # iterate columns of J22 = dS_dVm.imag at positions in pq (=pvpq)
        for c in range(Yp[pvpq[r]], Yp[pvpq[r] + 1]):
            cc = pvpq_lookup[Yj[c]]
            if pvpq[cc] == Yj[c]:
                # entry found J21
                # J[r + lpvpq, cc] = dVa_x[c].imag
                Jx[nnz] = dVa_x[c].imag
                Jj[nnz] = cc
                nnz += 1
                # also entry in J22
                Jx[nnz] = dVm_x[c].imag
                Jj[nnz] = cc + lpvpq
                nnz += 1
        # Jp: number of nonzeros per row = nnz - nnzStart (nnz at begging of loop - nnz at end of loop)
        Jp[r + lpvpq + 1] = nnz - nnzStart + Jp[r + lpvpq]


@jit(nopython=True, cache=False)
def create_J_ds(dVm_x, dVa_x, Yp, Yj, pvpq_lookup, refpvpq, pvpq, pq, Jx, Jj, Jp, slack_weights):  # pragma: no cover
    """Calculates Jacobian faster with numba and sparse matrices.

        Input: dS_dVa and dS_dVm in CSR sparse form (Yx = data, Yp = indptr, Yj = indices), pvpq, pq from pypower

        OUTPUT:  data from CSR form of Jacobian (Jx, Jj, Jp) and number of non zeros (nnz)

        @author: Roman Bolgaryn

        Calculate Jacobian entries

        J11 = dS_dVa[array([pvpq]).T, pvpq].real
        J12 = dS_dVm[array([pvpq]).T, pq].real
        J13 = slack_weights.reshape(-1,1)
        J21 = dS_dVa[array([pq]).T, pvpq].imag
        J22 = dS_dVm[array([pq]).T, pq].imag
        J23 = zeros(shape=(len(pq), 1))

        J has the shape
        | J11 | J12 | J13 |               | (ref+pvpq, pvpq) | (ref+pvpq, pq) | (ref+pvpq, 1) |
        | --------------- | = dimensions: | ------------------------------------------------- |
        | J21 | J22 | J23 |               |    (pq, pvpq)    |    (pq, pq)    |    (pq, 1)    |

        In addition to CreateJ, we add a row and a column
        We first iterate the rows of J11 and J12 (for r in range lpvpq) and add the entries which are stored in dS_dV
        We add the entries for slack weights (J13) if we encounter a positive slack weight in the row while iterating through J11 and J12
        Then we iterate the rows of J21 and J22 (for r in range lpq) and add the entries from dS_d
        The entries of J23 are always 0, so we do not do anything for J23 here (will be 0 because of the sparse matrix)

    """
    # Jacobi Matrix in sparse form
    # Jp, Jx, Jj equal J like:
    # J = zeros(shape=(ndim, ndim), dtype=float64)

    # Yp = Ybus.indptr (maps the elements of data and indices to the rows of the sparse matrix)
    # Yj = Ybus.indices (array mapping each element in data to its column in the sparse matrix)

    # get length of vectors
    lrefpvpq = len(refpvpq)
    lpvpq = len(pvpq)
    lpq = len(pq)
    lpv = lpvpq - lpq
    lrefpv = lrefpvpq - lpq
    lref = lrefpv - lpv

    # dense J matrix for debugging
    # if DEBUG:
    #     JJ = zeros(shape=(lrefpvpq+lpq, lrefpvpq+lpq))

    # nonzeros in J
    nnz = 0
    # iterate rows of J
    # first iterate pvpq (J11 and J12) -> the real part
    for r in range(lrefpvpq):
        # nnzStar is necessary to calculate nonzeros per row
        nnzStart = nnz
        # add slack weights to the first column
        if slack_weights[refpvpq[r]] > 0:
            Jx[nnz] = slack_weights[refpvpq[r]]
            Jj[nnz] = 0
            # if DEBUG:
            #     JJ[r, 0] = Jx[nnz]
            nnz += 1
        # iterate columns of J11 = dS_dVa.real at positions in pvpq
        # check entries in row pvpq[r] of dS_dV
        # r_start, r_end = Yp[refpvpq[r]], Yp[refpvpq[r] + 1]
        # for c in range(r_start, r_end):
        for c in range(Yp[refpvpq[r]], Yp[refpvpq[r] + 1]):
            # check if column Yj is in pvpq
            bus_idx = Yj[c]
            cc = pvpq_lookup[bus_idx]
            # entries for J11 and J12
            lookup_idx = refpvpq[cc]
            # skip if in the "ref" columns:
            skip = False
            for refcol in range(lref): # this loop is way faster
                if lookup_idx == refpvpq[refcol]:
                    skip = True
                    break
            if lookup_idx == bus_idx and not skip:# and lookup_idx in pvpq:  # too slow
                # entry found
                # equals entry of J11: J[r,cc] = dVa_x[c].real
                col = cc
                Jx[nnz] = dVa_x[c].real
                Jj[nnz] = col
                # if DEBUG:
                #     JJ[r, col] = Jx[nnz]
                nnz += 1
                # if entry is found in the "pq part" of pvpq = add entry of J12
                if cc >= lrefpv:
                    col = cc + lpq
                    Jx[nnz] = dVm_x[c].real
                    Jj[nnz] = col
                    # if DEBUG:
                    #     JJ[r, col] = Jx[nnz]
                    nnz += 1
        # Jp: number of nonzeros per row = nnz - nnzStart (nnz at begging of loop - nnz at end of loop)
        Jp[r + 1] = nnz - nnzStart + Jp[r]

    # second: iterate pq (J21 and J22) -> the imaginary part
    for r in range(lpq):
        nnzStart = nnz
        # iterate columns of J21 = dS_dVa.imag at positions in pvpq
        # r_start, r_end = Yp[pq[r]], Yp[pq[r] + 1]
        # for c in range(r_start, r_end):
        for c in range(Yp[pq[r]], Yp[pq[r] + 1]):
            bus_idx = Yj[c]
            cc = pvpq_lookup[bus_idx]
            lookup_idx = refpvpq[cc]
            skip = False
            for refcol in range(lref): # this loop is way faster
                if lookup_idx == refpvpq[refcol]:
                    skip = True
                    break
            if lookup_idx == bus_idx and not skip:# and lookup_idx in pvpq:   # too slow
                # entry found
                # equals entry of J21: J[r + lpvpq, cc] = dVa_x[c].imag
                col = cc
                Jx[nnz] = dVa_x[c].imag
                Jj[nnz] = col
                # if DEBUG:
                #     JJ[r+lrefpvpq, col] = Jx[nnz]
                nnz += 1
                # if entry is found in the "pq part" of pvpq = Add entry of J22
                if cc >= lrefpv:
                    col = cc + lpq
                    Jx[nnz] = dVm_x[c].imag
                    Jj[nnz] = col
                    # if DEBUG:
                    #     JJ[r+lrefpvpq, col] = Jx[nnz]
                    nnz += 1
        # Jp: number of nonzeros per row = nnz - nnzStart (nnz at begging of loop - nnz at end of loop)
        Jp[r + lrefpvpq + 1] = nnz - nnzStart + Jp[r + lrefpvpq]



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

from scipy.optimize import minimize

from pandapower.estimation.algorithm.base import BaseAlgorithm
from pandapower.estimation.algorithm.estimator import BaseEstimatorOpt, get_estimator
from pandapower.estimation.ppc_conversion import ExtendedPPCI

DEFAULT_OPT_METHOD = "Newton-CG"


# DEFAULT_OPT_METHOD = "TNC"
# DEFAULT_OPT_METHOD = "SLSQP"
# DEFAULT_OPT_METHOD = 'L-BFGS-B'


class OptAlgorithm(BaseAlgorithm):
    def estimate(self, eppci: ExtendedPPCI, estimator="wls", verbose=True, **kwargs):
        opt_method = DEFAULT_OPT_METHOD if 'opt_method' not in kwargs else kwargs['opt_method']

        # matrix calculation object
        estm = get_estimator(BaseEstimatorOpt, estimator)(eppci, **kwargs)

        jac = estm.create_cost_jacobian
        res = minimize(estm.cost_function, x0=eppci.E,
                       method=opt_method, jac=jac, tol=self.tolerance,
                       options={"disp": verbose})

        self.successful = res.success
        if self.successful:
            E = res.x
            eppci.update_E(E)
            return eppci
        else:
            raise Exception("Optimization failed! State Estimation not successful!")



import logging
import time

import pandas as pd

from pandapower.converter.cim import cim_tools
from pandapower.converter.cim.cim2pp import build_pp_net
from pandapower.converter.cim.other_classes import Report, LogLevel, ReportCode

logger = logging.getLogger('cim.cim2pp.converter_classes.conformLoadsCim16')

sc = cim_tools.get_pp_net_special_columns_dict()


class ConformLoadsCim16:
    def __init__(self, cimConverter: build_pp_net.CimConverter):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.cimConverter = cimConverter

    def convert_conform_loads_cim16(self):
        time_start = time.time()
        self.logger.info("Start converting ConformLoads.")
        eqssh_conform_loads = self._prepare_conform_loads_cim16()
        self.cimConverter.copy_to_pp('load', eqssh_conform_loads)
        self.logger.info("Created %s loads in %ss." % (eqssh_conform_loads.index.size, time.time() - time_start))
        self.cimConverter.report_container.add_log(Report(
            level=LogLevel.INFO, code=ReportCode.INFO_CONVERTING,
            message="Created %s loads from ConformLoads in %ss." %
                    (eqssh_conform_loads.index.size, time.time() - time_start)))

    def _prepare_conform_loads_cim16(self) -> pd.DataFrame:
        eqssh_conform_loads = self.cimConverter.merge_eq_ssh_profile('ConformLoad', add_cim_type_column=True)
        eqssh_conform_loads = pd.merge(eqssh_conform_loads, self.cimConverter.bus_merge, how='left', on='rdfId')
        if 'inService' in eqssh_conform_loads.columns:
            eqssh_conform_loads['connected'] = eqssh_conform_loads['connected'] & eqssh_conform_loads['inService']
        eqssh_conform_loads = eqssh_conform_loads.rename(columns={'rdfId': sc['o_id'], 'rdfId_Terminal': sc['t'], 'index_bus': 'bus',
                                            'connected': 'in_service', 'p': 'p_mw', 'q': 'q_mvar'})
        eqssh_conform_loads['const_i_percent'] = 0.
        eqssh_conform_loads['const_z_percent'] = 0.
        eqssh_conform_loads['scaling'] = 1.
        eqssh_conform_loads['type'] = None
        return eqssh_conform_loads



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import numpy as np
import pandas as pd

from packaging.version import Version

from pandapower._version import __version__, __format_version__
from pandapower.create import create_empty_network, create_poly_cost
from pandapower.results import reset_results
from pandapower.control import TrafoController
from pandapower.plotting.geo import convert_geodata_to_geojson

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


def convert_format(net, elements_to_deserialize=None):
    """
    Converts old nets to new format to ensure consistency. The converted net is returned.
    """
    from pandapower.toolbox import set_data_type_of_columns_to_default
    if not isinstance(net.version, str) or not hasattr(net, 'format_version') or \
            Version(net.format_version) > Version(net.version):
        net.format_version = net.version
    if isinstance(net.format_version, str) and Version(net.format_version) >= Version(__format_version__):
        return net
    _add_nominal_power(net)
    _add_missing_tables(net)
    _rename_columns(net, elements_to_deserialize)
    _add_missing_columns(net, elements_to_deserialize)
    _create_seperate_cost_tables(net, elements_to_deserialize)
    if Version(str(net.format_version)) < Version("3.0.0"):
        _convert_geo_data(net, elements_to_deserialize)
        _convert_group_element_index(net)
        _convert_trafo_controller_parameter_names(net)
        convert_trafo_pst_logic(net)
    if Version(str(net.format_version)) < Version("2.4.0"):
        _convert_bus_pq_meas_to_load_reference(net, elements_to_deserialize)
    if Version(str(net.format_version)) < Version("2.0.0"):
        _convert_to_generation_system(net, elements_to_deserialize)
        _convert_costs(net)
        _convert_to_mw(net)
        _update_trafo_parameter_names(net, elements_to_deserialize)
        reset_results(net)
    if Version(str(net.format_version)) < Version("1.6"):
        set_data_type_of_columns_to_default(net)
    _convert_objects(net, elements_to_deserialize)
    _update_characteristics(net, elements_to_deserialize)
    correct_dtypes(net, error=False)
    _add_missing_std_type_tables(net)
    net.format_version = __format_version__
    net.version = __version__
    _restore_index_names(net)
    return net


def _convert_geo_data(net, elements_to_deserialize=None):
    if ((_check_elements_to_deserialize('bus_geodata', elements_to_deserialize)
         and _check_elements_to_deserialize('bus', elements_to_deserialize))
        or (_check_elements_to_deserialize('line_geodata', elements_to_deserialize)
            and _check_elements_to_deserialize('line', elements_to_deserialize))):
        if hasattr(net, 'bus_geodata') or hasattr(net, 'line_geodata'):
            if Version(str(net.format_version)) < Version("1.6"):
                net.bus_geodata = pd.DataFrame.from_dict(net.bus_geodata)
                net.line_geodata = pd.DataFrame.from_dict(net.line_geodata)
            convert_geodata_to_geojson(net)


def _restore_index_names(net):
    """Restores dataframes index names stored as dictionary. With newer pp to_json() this
    information is stored to the dataframe its self.
    """
    if "index_names" in net.keys():
        if not isinstance(net["index_names"], dict):
            raise ValueError("To restore the index names of the dataframes, a dict including this "
                             f"information is expected, not {type(net['index_names'])}")
        for key, index_name in net["index_names"].items():
            if key in net.keys():
                net[key].index.name = index_name
        del net["index_names"]


def correct_dtypes(net, error):
    """
    Corrects all dtypes of pp element tables if possible. If not and error is True, an Error is
    raised.
    """
    empty_net = create_empty_network()
    empty_net.trafo['tap_changer_type'] = empty_net.trafo['tap_changer_type']
    empty_net.trafo3w['tap_changer_type'] = empty_net.trafo3w['tap_changer_type']
    not_corrected = list()
    failed = dict()
    for key, table in empty_net.items():
        if isinstance(table, pd.DataFrame):
            if key in net.keys() and isinstance(net[key], pd.DataFrame):
                cols = table.columns.intersection(net[key].columns)
                diff_cols = cols[~(table.dtypes.loc[cols] == net[key].dtypes.loc[cols])]
                for col in diff_cols:
                    try:
                        net[key][col] = net[key][col].astype(table[col].dtype)
                    except ValueError:
                        if key not in failed.keys():
                            failed[key] = [col]
                        else:
                            failed[key].append(col)
            else:
                not_corrected.append(key)
    if not_corrected:
        logger.warning("These keys were not corrected since they miss or are no dataframes: " + str(
            not_corrected))
    if failed:
        msg = "These dtypes could not be corrected: " + str(failed)
        if error:
            raise ValueError(msg)
        else:
            logger.info(msg)


def _convert_group_element_index(net):
    if isinstance(net.group, pd.DataFrame) and "element" in net.group.columns:
        if "element_index" in net.group.columns:
            logger.warning("element cannot be renamed by element_index because columns exist already.")
        net.group = net.group.rename(columns={"element": "element_index"})


def _convert_trafo_controller_parameter_names(net):
    if not isinstance(net.controller, pd.DataFrame):
        return
    for ctrl_idx in net.controller.index:
        controller = net.controller.at[ctrl_idx, "object"]
        if issubclass(type(controller), TrafoController):

            if "tid" in controller.__dict__.keys():
                controller.__dict__["element_index"] = controller.__dict__.pop("tid")
            elif "transformer_index" in controller.__dict__.keys():
                controller.__dict__["element_index"] = controller.__dict__.pop("transformer_index")

            if "trafotable" in controller.__dict__.keys():
                controller.__dict__["element"] = controller.__dict__.pop("trafotable")
                if "trafotype" in controller.__dict__.keys():
                    del controller.__dict__["trafotype"]
            elif "trafotype" in controller.__dict__.keys():
                controller.__dict__["element"] = controller.__dict__.pop("trafotype")

            if "controlled_bus" in controller.__dict__.keys():
                controller.__dict__["trafobus"] = controller.__dict__.pop("controlled_bus")


def _convert_bus_pq_meas_to_load_reference(net, elements_to_deserialize):
    if _check_elements_to_deserialize('measurement', elements_to_deserialize):
        bus_pq_meas_mask = net.measurement.measurement_type.isin(["p", "q"]) & \
                           (net.measurement.element_type == "bus")
        net.measurement.loc[bus_pq_meas_mask, "value"] *= -1


def _convert_to_generation_system(net, elements_to_deserialize):
    update_elements = []
    if _check_elements_to_deserialize('sgen', elements_to_deserialize):
        net.sgen.p_kw *= -1
        net.sgen.q_kvar *= -1
        update_elements += ['sgen']
    if _check_elements_to_deserialize('gen', elements_to_deserialize):
        net.gen.p_kw *= -1
        update_elements += ['gen']
    if _check_elements_to_deserialize('ext_grid', elements_to_deserialize):
        update_elements += ['ext_grid']
    for element in update_elements:
        for suffix in ["p_kw", "q_kvar"]:
            constraints = {}
            if "min_%s" % suffix in net[element]:
                constraints["max_%s" % suffix] = net[element]["min_%s" % suffix] * -1
                del net[element]["min_%s" % suffix]
            if "max_%s" % suffix in net[element]:
                constraints["min_%s" % suffix] = net[element]["max_%s" % suffix] * -1
                del net[element]["max_%s" % suffix]
            for column, values in constraints.items():
                net[element][column] = values
    pq_measurements = net.measurement[net.measurement.measurement_type.isin(["p", "q"])].index
    net.measurement.loc[pq_measurements, ["value", "std_dev"]] *= 1e-3


def _convert_costs(net):
    if "polynomial_cost" in net:
        for cost in net.polynomial_cost.itertuples():
            values = cost.c[0]
            if len(values) == 2:
                cp0 = values[1]
                cp1 = values[0]
                cp2 = 0
            elif len(values) == 3:
                cp0 = values[2]
                cp1 = values[1]
                cp2 = values[0]
            create_poly_cost(net, et=cost.element_type, element=cost.element, cp0_eur=cp0,
                             cp1_eur_per_mw=cp1 * 1e3, cp2_eur_per_mw2=cp2 * 1e6)
        del net.polynomial_cost
    if "piecewise_linear_cost" in net:
        if len(net.piecewise_linear_cost) > 0:
            raise NotImplementedError
        del net.piecewise_linear_cost


def _add_nominal_power(net):
    if "sn_kva" in net:
        net.sn_mva = net.pop("sn_kva") * 1e-3

    # Reset sn_mva only if sn_mva not available
    if "sn_mva" not in net:
        net.sn_mva = 1.0


def _add_missing_tables(net):
    net_new = create_empty_network()
    for key in net_new.keys():
        if key.startswith("_empty_res"):
            net[key] = net_new[key]
        elif key not in net.keys():
            net[key] = net_new[key]


def _create_seperate_cost_tables(net, elements_to_deserialize):
    if _check_elements_to_deserialize('gen', elements_to_deserialize) and "cost_per_kw" in net.gen:
        for index, cost in net.gen.cost_per_kw.items():
            if not np.isnan(cost):
                create_poly_cost(net, index, "gen", cp1_eur_per_mw=cost * 1e3)

    if _check_elements_to_deserialize('sgen', elements_to_deserialize) and \
            "cost_per_kw" in net.sgen:
        for index, cost in net.sgen.cost_per_kw.items():
            if not np.isnan(cost):
                create_poly_cost(net, index, "sgen", cp1_eur_per_kw=cost)

    if _check_elements_to_deserialize('ext_grid', elements_to_deserialize) and \
            "cost_per_kw" in net.ext_grid:
        for index, cost in net.ext_grid.cost_per_kw.items():
            if not np.isnan(cost):
                create_poly_cost(net, index, "ext_grid", cp1_eur_per_kw=cost)

    if _check_elements_to_deserialize('gen', elements_to_deserialize) and \
            "cost_per_kvar" in net.gen:
        for index, cost in net.gen.cost_per_kvar.items():
            if not np.isnan(cost):
                create_poly_cost(net, index, "ext_grid", cp1_eur_per_mw=0,
                                 cq1_eur_per_mvar=cost * 1e3)

    if _check_elements_to_deserialize('sgen', elements_to_deserialize) and \
            "cost_per_kvar" in net.sgen:
        for index, cost in net.sgen.cost_per_kvar.items():
            if not np.isnan(cost):
                create_poly_cost(net, index, "sgen", cp1_eur_per_mw=0,
                                 cq1_eur_per_mvar=cost * 1e3)

    if _check_elements_to_deserialize('ext_grid', elements_to_deserialize) and \
            "cost_per_kvar" in net.ext_grid:
        for index, cost in net.ext_grid.cost_per_kvar.items():
            if not np.isnan(cost):
                create_poly_cost(net, index, "ext_grid", cp1_eur_per_mw=0,
                                 cq1_eur_per_mvar=cost * 1e3)


def _rename_columns(net, elements_to_deserialize):
    if _check_elements_to_deserialize('line', elements_to_deserialize):
        net.line = net.line.rename(columns={'imax_ka': 'max_i_ka'})
    if _check_elements_to_deserialize('gen', elements_to_deserialize):
        net.gen = net.gen.rename(columns={"qmin_mvar": "min_q_mvar", "qmax_mvar": "max_q_mvar"})
    for typ, data in net.std_types["line"].items():
        if "imax_ka" in data:
            net.std_types["line"][typ]["max_i_ka"] = net.std_types["line"][typ].pop("imax_ka")
    _update_trafo_parameter_names(net, elements_to_deserialize)
    # initialize measurement dataframe
    if _check_elements_to_deserialize('measurement', elements_to_deserialize):
        if "measurement" in net and "type" in net.measurement and "measurement":
            if net.measurement.empty:
                net["measurement"] = create_empty_network()["measurement"]
            else:
                net.measurement["side"] = None
                bus_measurements = net.measurement.element_type == "bus"
                net.measurement.loc[bus_measurements, "element"] = \
                    net.measurement.loc[bus_measurements, "bus"].values
                net.measurement.loc[~bus_measurements, "side"] = \
                    net.measurement.loc[~bus_measurements, "bus"].values
                net.measurement = net.measurement.rename(columns={'type': 'measurement_type'})
                net.measurement = net.measurement.drop(["bus"], axis=1)
    if _check_elements_to_deserialize('controller', elements_to_deserialize):
        if "controller" in net:
            net["controller"] = net["controller"].rename(columns={"controller": "object"})
    if "options" in net:
        if "recycle" in net["options"]:
            if "Ybus" in net["options"]["recycle"]:
                if net["options"]["recycle"]["Ybus"]:
                    net["options"]["recycle"]["trafo"] = False
                del net["options"]["recycle"]["Ybus"]
            else:
                net["options"]["recycle"]["trafo"] = True
            if "ppc" in net["options"]["recycle"]:
                if net["options"]["recycle"]["ppc"]:
                    net["options"]["recycle"]["bus_pq"] = False
                del net["options"]["recycle"]["ppc"]
            else:
                net["options"]["recycle"]["bus_pq"] = True


def _add_missing_columns(net, elements_to_deserialize):
    update_elements = []
    if _check_elements_to_deserialize('trafo', elements_to_deserialize):
        update_elements += ['trafo']
    if _check_elements_to_deserialize('line', elements_to_deserialize):
        update_elements += ['line']
    for element in update_elements:
        if "df" not in net[element]:
            net[element]["df"] = 1.0

    if _check_elements_to_deserialize('bus', elements_to_deserialize) \
            and _check_elements_to_deserialize('bus_geodata', elements_to_deserialize) \
            and "geo" not in net.bus:
        net.bus["geo"] = np.nan
    if _check_elements_to_deserialize('trafo3w', elements_to_deserialize) and \
            "tap_at_star_point" not in net.trafo3w:
        net.trafo3w["tap_at_star_point"] = False
    if _check_elements_to_deserialize('trafo3w', elements_to_deserialize) and \
            "tap_step_degree" not in net.trafo3w:
        net.trafo3w["tap_step_degree"] = 0
    if _check_elements_to_deserialize('load', elements_to_deserialize) and \
            "const_z_percent" not in net.load or "const_i_percent" not in net.load:
        net.load["const_z_percent"] = np.zeros(net.load.shape[0])
        net.load["const_i_percent"] = np.zeros(net.load.shape[0])

    if _check_elements_to_deserialize('shunt', elements_to_deserialize) and \
            "vn_kv" not in net["shunt"]:
        net.shunt["vn_kv"] = net.bus.vn_kv.loc[net.shunt.bus.values].values
    if _check_elements_to_deserialize('shunt', elements_to_deserialize) and \
            "step" not in net["shunt"]:
        net.shunt["step"] = 1
    if _check_elements_to_deserialize('shunt', elements_to_deserialize) and \
            "max_step" not in net["shunt"]:
        net.shunt["max_step"] = 1
    if _check_elements_to_deserialize('trafo3w', elements_to_deserialize) and \
            "std_type" not in net.trafo3w:
        net.trafo3w["std_type"] = None

    if _check_elements_to_deserialize('sgen', elements_to_deserialize) and \
            "current_source" not in net.sgen:
        net.sgen["current_source"] = net.sgen["type"].apply(
            func=lambda x: False if x == "motor" else True)

    if _check_elements_to_deserialize('line', elements_to_deserialize):
        if "g_us_per_km" not in net.line:
            net.line["g_us_per_km"] = 0.
        if _check_elements_to_deserialize('line_geodata', elements_to_deserialize) and "geo" not in net.line:
            net.line["geo"] = np.nan

    if _check_elements_to_deserialize('gen', elements_to_deserialize) and \
            "slack" not in net.gen:
        net.gen["slack"] = False

    if _check_elements_to_deserialize('trafo', elements_to_deserialize) and \
            "tap_changer_type" not in net.trafo:
        net.trafo["tap_changer_type"] = None

    # asymmetric impedance
    if _check_elements_to_deserialize('impedance', elements_to_deserialize):
        if "r_pu" in net.impedance:
            net.impedance["rft_pu"] = net.impedance["rtf_pu"] = net.impedance["r_pu"]
            net.impedance["xft_pu"] = net.impedance["xtf_pu"] = net.impedance["x_pu"]
        if "gf_pu" not in net.impedance:
            net.impedance["gf_pu"] = 0.
            net.impedance["gt_pu"] = 0.
            net.impedance["bf_pu"] = 0.
            net.impedance["bt_pu"] = 0.

    # Update the switch table with 'z_ohm'
    if _check_elements_to_deserialize('switch', elements_to_deserialize) and \
            'z_ohm' not in net.switch:
        net.switch['z_ohm'] = 0

    # Update the switch table with 'in_ka'
    if _check_elements_to_deserialize('switch', elements_to_deserialize) and \
            'in_ka' not in net.switch:
        net.switch['in_ka'] = np.nan

    # Update the switch table with 'in_ka'
    if _check_elements_to_deserialize('res_switch', elements_to_deserialize) and \
            'p_from_mw' not in net.res_switch:
        net.res_switch['p_from_mw'] = np.nan
        net.res_switch['q_from_mvar'] = np.nan
        net.res_switch['p_to_mw'] = np.nan
        net.res_switch['q_to_mvar'] = np.nan

    # Update the switch table with 'in_ka'
    if _check_elements_to_deserialize('res_switch_est', elements_to_deserialize) and \
            'p_from_mw' not in net.res_switch_est:
        net.res_switch_est['p_from_mw'] = np.nan
        net.res_switch_est['q_from_mvar'] = np.nan
        net.res_switch_est['p_to_mw'] = np.nan
        net.res_switch_est['q_to_mvar'] = np.nan

    if _check_elements_to_deserialize('measurement', elements_to_deserialize) and \
            "name" not in net.measurement:
        net.measurement.insert(0, "name", None)

    if _check_elements_to_deserialize('controller', elements_to_deserialize) and \
            "initial_run" not in net.controller:
        net.controller.insert(4, 'initial_run', False)
        for _, ctrl in net.controller.iterrows():
            if hasattr(ctrl['object'], 'initial_run'):
                net.controller.at[ctrl.name, 'initial_run'] = ctrl['object'].initial_run
            else:
                net.controller.at[ctrl.name, 'initial_run'] = ctrl['object'].initial_powerflow

    # distributed slack
    if _check_elements_to_deserialize('ext_grid', elements_to_deserialize) and \
            "slack_weight" not in net.ext_grid:
        net.ext_grid['slack_weight'] = 1.0

    if _check_elements_to_deserialize('gen', elements_to_deserialize) and \
            "slack_weight" not in net.gen:
        net.gen['slack_weight'] = 0.0

    if _check_elements_to_deserialize('xward', elements_to_deserialize) and \
            "slack_weight" not in net.xward:
        net.xward['slack_weight'] = 0.0


def _update_trafo_type_parameter_names(net):
    for element in ('trafo', 'trafo3w'):
        for type in net.std_types[element].keys():
            keys = {col: _update_column(col) for col in net.std_types[element][type].keys() if
                    col.startswith("tp") or col.startswith("vsc")}
            for old_key, new_key in keys.items():
                net.std_types[element][type][new_key] = net.std_types[element][type].pop(old_key)


def _update_trafo_parameter_names(net, elements_to_deserialize):
    update_data = []
    if _check_elements_to_deserialize('trafo', elements_to_deserialize):
        update_data += ['trafo']
    if _check_elements_to_deserialize('trafo3w', elements_to_deserialize):
        update_data += ['trafo3w']
    for element in update_data:
        replace_cols = {col: _update_column(col) for col in net[element].columns if
                        col.startswith("tp") or col.startswith("vsc")}
        net[element] = net[element].rename(columns=replace_cols)
    _update_trafo_type_parameter_names(net)


def _update_column(column):
    column = column.replace("tp_", "tap_")
    column = column.replace("_st_", "_step_")
    column = column.replace("_mid", "_neutral")
    column = column.replace("vsc", "vk")
    return column


def _set_data_type_of_columns(net):
    new_net = create_empty_network()
    for key, item in net.items():
        if isinstance(item, pd.DataFrame):
            for col in item.columns:
                if col == "tap_pos":
                    continue
                if key in new_net and col in new_net[key].columns:
                    if set(item.columns) == set(new_net[key]):
                        net[key] = net[key].reindex(new_net[key].columns, axis=1)
                    net[key][col] = net[key][col].astype(new_net[key][col].dtype,
                                                         errors="ignore")


def _convert_to_mw(net):
    replace = [("kw", "mw"), ("kvar", "mvar"), ("kva", "mva")]
    for element in net.keys():
        if isinstance(net[element], pd.DataFrame):
            for old, new in replace:
                diff = {column: column.replace(old, new) for column in net[element].columns if
                        old in column and column != "pfe_kw"}
                net[element] = net[element].rename(columns=diff)
                if len(net[element]) == 0:
                    continue
                for old, new in diff.items():
                    net[element][new] *= 1e-3

    for element, std_types in net.std_types.items():
        for std_type, parameters in std_types.items():
            for parameter in set(parameters.keys()):
                for old, new in replace:
                    if old in parameter and parameter != "pfe_kw":
                        parameters[parameter.replace(old, new)] = parameters[parameter] * 1e-3
                        del parameters[parameter]


def _update_object_attributes(obj):
    """
    Rename attributes of a given object. A new attribute is added and the old one is removed.
    """
    to_rename = {"u_set": "vm_set_pu",
                 "u_lower": "vm_lower_pu",
                 "u_upper": "vm_upper_pu"}

    for key, val in to_rename.items():
        if key in obj.__dict__:
            obj.__dict__[val] = obj.__dict__.pop(key)

    if "vm_lower_pu" in obj.__dict__ and "hunting_limit" not in obj.__dict__:
        obj.__dict__["hunting_limit"] = None


def _convert_objects(net, elements_to_deserialize):
    """
    The function updates attribute names in pandapower objects. For now, it affects TrafoController.
    Should be expanded for other objects if necessary.
    """
    _check_elements_to_deserialize('controller', elements_to_deserialize)
    if _check_elements_to_deserialize('controller', elements_to_deserialize) and \
            "controller" in net.keys():
        for obj in net["controller"].object.values:
            _update_object_attributes(obj)


def _check_elements_to_deserialize(element, elements_to_deserialize):
    if elements_to_deserialize is None:
        return True
    else:
        return element in elements_to_deserialize


def _add_missing_std_type_tables(net):
    type_names = ("fuse", "line_dc")
    for tn in type_names:
        if tn not in net.std_types:
            net.std_types[tn] = {}


def _update_characteristics(net, elements_to_deserialize):
    # new interpolator type has been added to SplineCharacteristic - "pchip", and the attributes have been refactored
    if not _check_elements_to_deserialize("characteristic", elements_to_deserialize) or \
            "characteristic" not in net or net.characteristic.empty:
        return
    for c in net.characteristic.object.values:
        # meta check for old SplineCharacteristic (cannot import it here to use isinstance):
        if not (hasattr(c, "kind") and hasattr(c, "fill_value")):
            continue
        c.interpolator_kind = "interp1d"
        c.kwargs = {"kind": c.__dict__.pop("kind"), "bounds_error": False, "fill_value": c.__dict__.pop("fill_value")}


def convert_trafo_pst_logic(net):
    """
    Converts trafo and trafo3w phase shifter logic to version 3.0 or later
    """
    for trafotable in ["trafo", "trafo3w"]:
        if trafotable in net and isinstance(net[trafotable], pd.DataFrame):
            if net[trafotable].index.size > 0:
                for t in ("", "2"):
                    # drop old tap_phase_shifter flag
                    if f"tap{t}_phase_shifter" in net[trafotable]:
                        net[trafotable] = net[trafotable].drop(columns=f"tap{t}_phase_shifter")
                    if (f"tap{t}_step_degree" in net[trafotable]) or (f"tap{t}_step_percent" in net[trafotable]):
                        # no phase shifters - check if both tap_step_percent & tap_step_degree are 0 or nan
                        mask_na = (((net[trafotable][f"tap{t}_step_degree"].isna()) |
                                   (net[trafotable][f"tap{t}_step_degree"] == 0)) &
                                   ((net[trafotable][f"tap{t}_step_percent"].isna()) |
                                    (net[trafotable][f"tap{t}_step_percent"] == 0)))
                        net[trafotable].loc[mask_na, f"tap{t}_changer_type"] = None
                        # ratio/asymmetrical phase shifters
                        mask_ratio_asym = ((net[trafotable][f"tap{t}_step_degree"] != 90) &
                                           ((net[trafotable][f"tap{t}_step_percent"].notna()) &
                                            (net[trafotable][f"tap{t}_step_percent"] != 0)))
                        net[trafotable].loc[mask_ratio_asym, f"tap{t}_changer_type"] = "Ratio"
                        # symmetrical phase shifters
                        mask_sym = ((net[trafotable][f"tap{t}_step_degree"] == 90) &
                                    ((net[trafotable][f"tap{t}_step_percent"].notna()) &
                                    (net[trafotable][f"tap{t}_step_percent"] != 0)))
                        net[trafotable].loc[mask_sym, f"tap{t}_changer_type"] = "Symmetrical"
                        # ideal phase shifters
                        mask_ideal = (((net[trafotable][f"tap{t}_step_degree"].notna()) &
                                      (net[trafotable][f"tap{t}_step_degree"] != 0)) &
                                      ((net[trafotable][f"tap{t}_step_percent"].isna()) |
                                       (net[trafotable][f"tap{t}_step_percent"] == 0)))
                        net[trafotable].loc[mask_ideal, f"tap{t}_changer_type"] = "Ideal"



import time
from pandapower.grid_equivalents.auxiliary import drop_internal_branch_elements, \
    _runpp_except_voltage_angles, impedance_columns
import pandas as pd
import numpy as np

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

from pandapower.create import create_shunt, create_impedance, create_switch, create_ward, create_xward
from pandapower.toolbox.grid_modification import drop_buses

logger = logging.getLogger(__name__)


def _calculate_ward_and_impedance_parameters(Ybus_eq, bus_lookups, show_computing_time, power_eq=0):
    """calculates the wards and equivalente impedance to represente the
    external network"""
    t_start = time.perf_counter()
    # --- calculate ward paramter
    b_buses_ppc = bus_lookups["bus_lookup_ppc"]["b_area_buses"]
    b_buses_pd = bus_lookups["bus_lookup_pd"]["b_area_buses"]
    nb_b_buses_ppc = len(b_buses_ppc)

    ward_parameter = pd.DataFrame(columns=["bus_pd", "bus_ppc", "shunt", "power_eq"])
    ward_parameter["bus_ppc"] = b_buses_ppc
    ward_parameter["bus_pd"] = b_buses_pd
    ward_parameter["shunt"] = Ybus_eq.sum(axis=1)[-nb_b_buses_ppc:]
    ward_parameter["power_eq"] = 0 + 1j * 0  # power_eq.power_eq.values

    # --- calculate impedance paramter
    params = Ybus_eq[-nb_b_buses_ppc:, -nb_b_buses_ppc:]
    nl = nb_b_buses_ppc * (nb_b_buses_ppc - 1) // 2
    impedance_parameter = pd.DataFrame(
        data=np.arange(nl * len(impedance_columns)).reshape((nl, len(impedance_columns))), columns=impedance_columns,
        dtype=np.float64)
    k = 0
    for i in range(nb_b_buses_ppc):
        for j in range(nb_b_buses_ppc):
            if j > i:
                if np.abs(params[i, j]) > 1e-10:
                    impedance_parameter.loc[k, 'from_bus'] = b_buses_pd[i]
                    impedance_parameter.loc[k, 'to_bus'] = b_buses_pd[j]
                    impedance_parameter.loc[k, 'rft_pu'] = (-1 / params[i, j]).real
                    impedance_parameter.loc[k, 'xft_pu'] = (-1 / params[i, j]).imag
                    impedance_parameter.loc[k, 'rtf_pu'] = (-1 / params[j, i]).real
                    impedance_parameter.loc[k, 'xtf_pu'] = (-1 / params[j, i]).imag
                    k += 1
                else:
                    impedance_parameter = impedance_parameter[:-1]
    t_end = time.perf_counter()
    if show_computing_time:
        logger.info("\"calculate_ward_and_impedance_parameters\" finished in %s seconds:" % round((
                t_end - t_start), 2))
    return ward_parameter, impedance_parameter


def _calculate_xward_and_impedance_parameters(net_external, Ybus_eq, bus_lookups,
                                              show_computing_time, power_eq=0):
    """calculates the xwards and the equivalent impedance"""
    t_start = time.perf_counter()
    xward_parameter, impedance_parameter = \
        _calculate_ward_and_impedance_parameters(Ybus_eq, bus_lookups, False)
    xward_parameter["r_ohm"] = 0
    xward_parameter["x_ohm"] = -1 / xward_parameter.shunt.values.imag / \
                               net_external.sn_mva * net_external.bus.vn_kv[xward_parameter.bus_pd].values ** 2  #/2
    # np.square(net_external.bus.vn_kv[xward_parameter.bus_pd.values].values) / \
    # net_external.sn_mva/2
    xward_parameter["vm_pu"] = net_external.res_bus.vm_pu[xward_parameter.bus_pd.values].values
    t_end = time.perf_counter()
    if show_computing_time:
        logger.info("\"calculate_xward_and_impedance_parameters\" finished in %s seconds:" % round((
                t_end - t_start), 2))
    return xward_parameter, impedance_parameter


def create_passive_external_net_for_ward_admittance(
        net, all_external_buses, boundary_buses, calc_volt_angles=True,
        runpp_fct=_runpp_except_voltage_angles, **kwargs):
    """
    This function replace the wards and xward in external network by internal
    elements, and replace the power injections in external area by shunts
    if necessary.

    INPUT:
        **net** - The pandapower format network

        **all_external_buses** (list) -  list of all external buses

        **boundary_buses** (list) -  list of boundary bus indices, by which the
            original network are divide into an internal area and an external
            area
    """
    # --- replace power injection in external net by shunts to creat a passiv network
    v_m = net.res_bus.vm_pu[all_external_buses].values
    current_injections = (net.res_bus.p_mw[all_external_buses].values -
                          1j * net.res_bus.q_mvar[all_external_buses].values)
    shunt_params = list(current_injections / v_m ** 2)
    # creats shunts
    for i in range(len(all_external_buses)):
        if abs(np.nan_to_num(shunt_params[i])) != 0:
            create_shunt(net, all_external_buses[i], -shunt_params[i].imag,
                         shunt_params[i].real)
    # drops all power injections
    for elm in ["sgen", "gen", "load", "storage"]:
        target_idx = net[elm].index[net[elm].bus.isin(all_external_buses)]
        net[elm] = net[elm].drop(target_idx)
    runpp_fct(net, calculate_voltage_angles=calc_volt_angles, **kwargs)


def _replace_external_area_by_wards(net_external, bus_lookups, ward_parameter_no_power,
                                    impedance_parameter, ext_buses_with_xward,
                                    show_computing_time, calc_volt_angles=True,
                                    runpp_fct=_runpp_except_voltage_angles, **kwargs):
    t_start = time.perf_counter()
    """replaces the external networks by wards and equivalent impedance"""
    # --- drop all external elements
    e_buses_pd = bus_lookups["bus_lookup_pd"]["e_area_buses"]
    drop_buses(net_external, e_buses_pd)
    drop_internal_branch_elements(net_external, bus_lookups["boundary_buses_inclusive_bswitch"])
    #    runpp_fct(net_external, calculate_voltage_angles=True, **kwargs)

    # --- drop shunt elements attached to boundary buses
    traget_shunt_idx = net_external.shunt.index[net_external.shunt.bus.isin(bus_lookups[
                                                                                "boundary_buses_inclusive_bswitch"])]
    net_external.shunt = net_external.shunt.drop(traget_shunt_idx)

    # --- creat impedance
    sn = net_external.sn_mva
    for idx in impedance_parameter.index:
        from_bus = impedance_parameter.from_bus[idx]
        to_bus = impedance_parameter.to_bus[idx]
        if abs(impedance_parameter.rft_pu[idx]) > 1e-8 or \
                abs(impedance_parameter.xft_pu[idx]) > 1e-8 or \
                abs(impedance_parameter.rtf_pu[idx]) > 1e-8 or \
                abs(impedance_parameter.xtf_pu[idx]) > 1e-8:
            create_impedance(net_external, from_bus, to_bus,
                             impedance_parameter.rft_pu[idx],
                             impedance_parameter.xft_pu[idx],
                             sn_mva=sn,
                             rtf_pu=impedance_parameter.rtf_pu[idx],
                             xtf_pu=impedance_parameter.xtf_pu[idx],
                             name="eq_impedance")
        else:
            create_switch(net_external, from_bus, to_bus, "b", name="eq_switch")

    # --- creata ward
    for i in ward_parameter_no_power.index:
        target_bus = ward_parameter_no_power.bus_pd[i]
        create_ward(net_external, target_bus,
                    0.0,  # np.nan_to_num(-ward_parameter.power_eq[i].real),
                    0.0,  # np.nan_to_num(-ward_parameter.power_eq[i].imag),
                    ward_parameter_no_power.shunt[i].real * sn / (net_external.res_bus.vm_pu[target_bus] ** 2),
                    -ward_parameter_no_power.shunt[i].imag * sn / (net_external.res_bus.vm_pu[target_bus] ** 2),
                    name="network_equivalent")

    eq_power = net_external.res_ext_grid.copy()
    eq_power["bus"] = net_external.ext_grid.bus.values
    eq_power["elm"] = "ext_grid"
    slack_gen = net_external.gen.index[net_external.gen.slack]
    if len(slack_gen) != 0:
        for i in slack_gen:
            new_eq_power = \
                [net_external.res_gen.p_mw[i], net_external.res_gen.q_mvar[i], \
                 net_external.gen.bus[i], "gen"]
            eq_power.loc[len(eq_power)] = new_eq_power
    assert len(eq_power.bus) == len(set(eq_power.bus))  # only one slack at individual bus

    runpp_fct(net_external, calculate_voltage_angles=calc_volt_angles, **kwargs)

    eq_power.p_mw -= \
        pd.concat([net_external.res_ext_grid.p_mw, net_external.res_gen.p_mw[slack_gen]])
    eq_power.q_mvar -= \
        pd.concat([net_external.res_ext_grid.q_mvar, net_external.res_gen.q_mvar[slack_gen]])
    for bus in eq_power.bus:
        net_external.ward.loc[net_external.ward.bus == bus, 'ps_mw'] = \
            eq_power.p_mw[eq_power.bus == bus].values
        net_external.ward.loc[net_external.ward.bus == bus, 'qs_mvar'] = \
            eq_power.q_mvar[eq_power.bus == bus].values

    net_external.poly_cost = net_external.poly_cost[0:0]
    net_external.pwl_cost = net_external.pwl_cost[0:0]
    if len(ext_buses_with_xward):
        drop_buses(net_external,
                   net_external.bus.index.tolist()[-(len(ext_buses_with_xward)):])
    # net_external.ward.qs_mvar[i] = eq_power.q_mvar[
    #     net_external.ext_grid.bus == ward_parameter_no_power.bus_pd[i]]
    t_end = time.perf_counter()
    if show_computing_time:
        logger.info("\"replace_external_area_by_wards\" finished in %s seconds:" % round((
                t_end - t_start), 2))


def _replace_external_area_by_xwards(net_external, bus_lookups, xward_parameter_no_power,
                                     impedance_parameter, ext_buses_with_xward,
                                     show_computing_time, calc_volt_angles=True,
                                     runpp_fct=_runpp_except_voltage_angles, **kwargs):
    """replaces the external networks by xwards and equivalent impedance"""
    t_start = time.perf_counter()
    # --- drop all external elements
    e_buses_pd = bus_lookups["bus_lookup_pd"]["e_area_buses"]
    drop_buses(net_external, e_buses_pd)
    drop_internal_branch_elements(net_external, bus_lookups["boundary_buses_inclusive_bswitch"])
    # --- drop shunt elements attached to boundary buses
    traget_shunt_idx = net_external.shunt.index[net_external.shunt.bus.isin(bus_lookups[
                                                                                "boundary_buses_inclusive_bswitch"])]
    net_external.shunt = net_external.shunt.drop(traget_shunt_idx)

    # --- creat impedance
    sn = net_external.sn_mva
    for idx in impedance_parameter.index:
        from_bus = impedance_parameter.from_bus[idx]
        to_bus = impedance_parameter.to_bus[idx]
        if abs(impedance_parameter.rft_pu[idx]) > 1e-8 or \
                abs(impedance_parameter.xft_pu[idx]) > 1e-8 or \
                abs(impedance_parameter.rtf_pu[idx]) > 1e-8 or \
                abs(impedance_parameter.xtf_pu[idx]) > 1e-8:
            create_impedance(net_external, from_bus, to_bus,
                             impedance_parameter.rft_pu[idx],
                             impedance_parameter.xft_pu[idx],
                             sn_mva=net_external.sn_mva,
                             rtf_pu=impedance_parameter.rtf_pu[idx],
                             xtf_pu=impedance_parameter.xtf_pu[idx],
                             name="eq_impedance")
        else:
            create_switch(net_external, from_bus, to_bus, "b", name="eq_switch")
    # --- creata xward
    for i in xward_parameter_no_power.index:
        target_bus = xward_parameter_no_power.bus_pd[i]
        create_xward(net_external, target_bus,
                     0.0,  # np.nan_to_num(-xward_parameter.power_eq[i].real),
                     0.0,  # np.nan_to_num(-xward_parameter.power_eq[i].imag),
                     xward_parameter_no_power.shunt[i].real * sn / xward_parameter_no_power.vm_pu[i] ** 2,
                     0.0,
                     xward_parameter_no_power.r_ohm[i],
                     np.nan_to_num(xward_parameter_no_power.x_ohm[i]),
                     # neginf=1e100 is commented since this led to error
                     xward_parameter_no_power.vm_pu[i],
                     name="network_equivalent")

    eq_power = net_external.res_ext_grid.copy()
    eq_power["bus"] = net_external.ext_grid.bus.values
    eq_power["elm"] = "ext_grid"
    slack_gen = net_external.gen.index[net_external.gen.slack]
    if len(slack_gen) != 0:
        for i in slack_gen:
            new_eq_power = \
                [net_external.res_gen.p_mw[i], net_external.res_gen.q_mvar[i], \
                 net_external.gen.bus[i], "gen"]
            eq_power.loc[len(eq_power)] = new_eq_power
    assert len(eq_power.bus) == len(set(eq_power.bus))  # only one slack at individual bus

    runpp_fct(net_external, calculate_voltage_angles=calc_volt_angles,
              tolerance_mva=1e-6, max_iteration=100, **kwargs)

    eq_power.p_mw -= pd.concat([net_external.res_ext_grid.p_mw, net_external.res_gen.p_mw[slack_gen]])
    eq_power.q_mvar -= pd.concat([net_external.res_ext_grid.q_mvar, net_external.res_gen.q_mvar[slack_gen]])
    for bus in eq_power.bus:
        net_external.xward.loc[net_external.xward.bus == bus, "ps_mw"] =  eq_power.p_mw[eq_power.bus == bus].values
        net_external.xward.loc[net_external.xward.bus == bus, "qs_mvar"] = eq_power.q_mvar[eq_power.bus == bus].values

    net_external.poly_cost = net_external.poly_cost[0:0]
    net_external.pwl_cost = net_external.pwl_cost[0:0]
    if len(ext_buses_with_xward):
        drop_buses(net_external, net_external.bus.index.tolist()[-(len(ext_buses_with_xward)):])
    t_end = time.perf_counter()
    if show_computing_time:
        logger.info("\"replace_external_area_by_xwards\" finished in %s seconds:" % round((t_end - t_start), 2))



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import numpy as np
import pandas as pd
from pandas import DataFrame, notnull, isnull

from pandapower.run import runpp
from pandapower.create import create_measurement
from pandapower.topology import create_nxgraph, connected_component


def estimate_voltage_vector(net):
    """
    Function initializes the voltage vector of net with a rough estimation. All buses are set to the
    slack bus voltage. Transformer differences in magnitude and phase shifting are accounted for.
    :param net: pandapower network
    :return: pandas dataframe with estimated vm_pu and va_degree
    """
    res_bus = DataFrame(index=net.bus.index, columns=["vm_pu", "va_degree"])
    net_graph = create_nxgraph(net, include_trafos=False)
    for _, ext_grid in net.ext_grid.iterrows():
        area = list(connected_component(net_graph, ext_grid.bus))
        res_bus.loc[area, "vm_pu"] = ext_grid.vm_pu
        res_bus.loc[area, "va_degree"] = ext_grid.va_degree
    trafos = net.trafo[net.trafo.in_service == 1]
    trafo_index = trafos.index.tolist()
    while len(trafo_index):
        for tix in trafo_index:
            trafo = trafos.loc[tix]
            if notnull(res_bus.vm_pu.at[trafo.hv_bus]) and isnull(res_bus.vm_pu.at[trafo.lv_bus]):
                try:
                    area = list(connected_component(net_graph, trafo.lv_bus))
                    shift = trafo.shift_degree if "shift_degree" in trafo else 0
                    ratio = (trafo.vn_hv_kv / trafo.vn_lv_kv) / (net.bus.vn_kv.at[trafo.hv_bus]
                                                                 / net.bus.vn_kv.at[trafo.lv_bus])
                    res_bus.loc[area, "vm_pu"] = res_bus.vm_pu.at[trafo.hv_bus] * ratio
                    res_bus.loc[area, "va_degree"] = res_bus.va_degree.at[trafo.hv_bus] - shift
                except KeyError:
                    raise UserWarning("An out-of-service bus is connected to an in-service "
                                      "transformer. Please set the transformer out of service or"
                                      "put the bus into service. Treat results with caution!")
                trafo_index.remove(tix)
            elif notnull(res_bus.vm_pu.at[trafo.hv_bus]):
                # parallel transformer, lv buses are already set from previous transformer
                trafo_index.remove(tix)
            if len(trafo_index) == len(trafos):
                # after the initial run we could not identify any areas correctly, it's probably a transmission grid
                # with slack on the LV bus and multiple transformers/gens. do flat init and return
                res_bus.loc[res_bus.vm_pu.isnull(), "vm_pu"] = 1.
                res_bus.loc[res_bus.va_degree.isnull(), "va_degree"] = 0.
                return res_bus
    return res_bus


def _get_bus_ppc_mapping(net, bus_to_be_fused):
    bus_with_elements = set(net.load.bus).union(set(net.sgen.bus)).union(
        set(net.shunt.bus)).union(set(net.gen.bus)).union(
        set(net.ext_grid.bus)).union(set(net.ward.bus)).union(
        set(net.xward.bus))
    # Run dc pp to get the ppc we need
    #rundcpp(net)

    runpp(net, calculate_voltage_angles=True)

    bus_ppci = pd.DataFrame(data=net._pd2ppc_lookups['bus'], columns=["bus_ppci"])
    bus_ppci['bus_with_elements'] = bus_ppci.index.isin(bus_with_elements)
    existed_bus = bus_ppci[bus_ppci.index.isin(net.bus.index)]
    bus_ppci['vn_kv'] = net.bus.loc[existed_bus.index, 'vn_kv']
    ppci_bus_with_elements = bus_ppci.groupby('bus_ppci')['bus_with_elements'].sum()
    bus_ppci.loc[:, 'elements_in_cluster'] = ppci_bus_with_elements[bus_ppci['bus_ppci'].values].values
    bus_ppci['bus_to_be_fused'] = False
    if bus_to_be_fused is not None:
        bus_ppci.loc[bus_to_be_fused, 'bus_to_be_fused'] = True
        bus_cluster_to_be_fused_mask = bus_ppci.groupby('bus_ppci')['bus_to_be_fused'].any()
        bus_ppci.loc[bus_cluster_to_be_fused_mask[bus_ppci['bus_ppci'].values].values, 'bus_to_be_fused'] = True
    return bus_ppci


def set_bb_switch_impedance(net, bus_to_be_fused=None, z_ohm=0.1):
    """
     Assuming a substation with multiple buses connected to each other with bb switch
     if multiple of them have an element (load, sgen usw.) then it caused problem on
     e.g. State Estimation, since they will be automatically fused in ppc
     A possiblity would be to properly seperate the substation to multiple clusters (ppc) bus
     according to the number of buses with elements, in order to estimate the p, q on all
     the busses.
     These method implemented a forward and backward substitution:
     1. Find the mapping of pd2ppc bus, and identify the ppc bus, which contains multiple bus with elements
     2. Skip the first bus with elements with the same ppc bus, and finding all the switch connected to the
       rest pandapower bus and set impedance
     3. Update the pd2ppc lookup
     4. if it caused some bus without elements to detach from any of the bus with connected element, then the
       switch caused it will be reset to fuse mode
     5. Iterate this process until all buses with elements are seperated to their own ppc bus, while no bus is
       detached in order to get the voltage on bus right
    :param net: pandapower net
    :param z_ohm: float z_ohm to be setted for the switch
    :return: None
    """
    if 'z_ohm' in net.switch:
        net.switch['z_ohm_ori'] = net.switch['z_ohm']

    lookup = _get_bus_ppc_mapping(net, bus_to_be_fused)
    for _ in range(int(lookup.elements_in_cluster.max()) - 1):
        bus_to_be_handled = lookup[((lookup['elements_in_cluster'] >= 2) &
                                    lookup['bus_with_elements']) & (~lookup['bus_to_be_fused'])]
        bus_to_be_handled = bus_to_be_handled[bus_to_be_handled['bus_ppci'].duplicated(keep='first')].index
        imp_switch_sel = (net.switch.et == 'b') & (net.switch.closed) &\
                         (net.switch.bus.isin(bus_to_be_handled) | net.switch.element.isin(bus_to_be_handled))
        net.switch.loc[imp_switch_sel, 'z_ohm'] = z_ohm

        # check if fused buses were isolated by the switching type change
        lookup_discon = _get_bus_ppc_mapping(net, bus_to_be_fused)
        imp_switch = net.switch.loc[imp_switch_sel, :]
        detached_bus = lookup_discon.loc[lookup_discon.elements_in_cluster == 0, :].index
        # Find the cause of isolated bus
        switch_to_be_fused = (imp_switch.bus.isin(bus_to_be_handled) & imp_switch.element.isin(detached_bus)) | \
                             (imp_switch.bus.isin(detached_bus) & imp_switch.element.isin(bus_to_be_handled))
        if not switch_to_be_fused.values.any():
            return
        net.switch.loc[imp_switch[switch_to_be_fused].index, 'z_ohm'] = 0
        lookup = _get_bus_ppc_mapping(net, bus_to_be_fused)
        if lookup.elements_in_cluster.max() == 1:
            return


def reset_bb_switch_impedance(net):
    """
    Reset the z_ohm of the switch to its original state, undo the operation by set_bb_switch_impedance
    :param net: pandapower net
    :return: None
    """
    if "z_ohm_ori" in net.switch:
        net.switch["z_ohm"] = net.switch["z_ohm_ori"]
        net.switch = net.switch.drop("z_ohm_ori", axis=1)


def add_virtual_meas_from_loadflow(net, v_std_dev=0.01, p_std_dev=0.03, q_std_dev=0.03,
                                   seed=14, with_random_error=False):
    np.random.seed(seed)

    bus_meas_types = {'v': 'vm_pu', 'p': 'p_mw', 'q': 'q_mvar'}
    branch_meas_type = {'line': {'side': ('from', 'to'),
                                 'meas_type': ('p_mw', 'q_mvar')},
                        'trafo': {'side': ('hv', 'lv'),
                                  'meas_type': ('p_mw', 'q_mvar')},
                        'trafo3w': {'side': ('hv', 'mv', 'lv'),
                                    'meas_type': ('p_mw', 'q_mvar')}}
    for bus_ix, bus_res in net.res_bus.iterrows():
        for meas_type in bus_meas_types.keys():
            meas_value = float(bus_res[bus_meas_types[meas_type]])
            if meas_type in ('p', 'q'):
                create_measurement(net, meas_type=meas_type, element_type='bus', element=bus_ix,
                                      value=meas_value, std_dev=1)
            else:
                create_measurement(net, meas_type=meas_type, element_type='bus', element=bus_ix,
                                      value=meas_value, std_dev=v_std_dev)
    remove_shunt_injection_from_meas(net,"shunt")
    remove_shunt_injection_from_meas(net,"ward")

    for br_type in branch_meas_type.keys():
        if not net['res_' + br_type].empty:
            for br_ix, br_res in net['res_' + br_type].iterrows():
                for side in branch_meas_type[br_type]['side']:
                    for meas_type in branch_meas_type[br_type]['meas_type']:
                        create_measurement(net, meas_type=meas_type[0], element_type=br_type,
                                              element=br_ix, side=side,
                                              value=br_res[meas_type[0] + '_' + side + meas_type[1:]], std_dev=1)

    add_virtual_meas_error(net, v_std_dev=v_std_dev, p_std_dev=p_std_dev, q_std_dev=q_std_dev,
                           with_random_error=with_random_error)

def remove_shunt_injection_from_meas(net,type):
    index = net[type].index.tolist()
    bus = net[type]["bus"].tolist()
    for k in range(len(index)):
        try:
            idxp = net.measurement[((net.measurement["element_type"]=="bus") & (net.measurement["element"]==bus[k])) & (net.measurement["measurement_type"]=="p")].index[0]
            idxq = net.measurement[((net.measurement["element_type"]=="bus") & (net.measurement["element"]==bus[k])) & (net.measurement["measurement_type"]=="q")].index[0]
            if type == "shunt":
                net.measurement.loc[idxp,"value"] -= net.res_shunt.loc[index[k], 'p_mw']
                net.measurement.loc[idxq,"value"] -= net.res_shunt.loc[index[k], 'q_mvar']
            if type == "ward":
                net.measurement.loc[idxp,"value"] -= net.res_ward.loc[index[k], 'p_mw'] - net.ward.loc[index[k], 'ps_mw']
                net.measurement.loc[idxq,"value"] -= net.res_ward.loc[index[k], 'q_mvar'] - net.ward.loc[index[k], 'qs_mvar']
        except:
            continue

def add_virtual_pmu_meas_from_loadflow(net, v_std_dev=0.001, i_std_dev=0.1,
                                       p_std_dev=0.01, q_std_dev=0.01, dg_std_dev=0.1,
                                       seed=14, with_random_error=True):
    np.random.seed(seed)

    bus_meas_types = {'v': 'vm_pu', "va": "va_degree", 'p': 'p_mw', 'q': 'q_mvar'}
    branch_meas_type = {'line': {'side': ('from', 'to'),
                                 'meas_type': ('i_ka', 'ia_degree', 'p_mw', 'q_mvar')},
                        'trafo': {'side': ('hv', 'lv'),
                                  'meas_type': ('i_ka', 'ia_degree', 'p_mw', 'q_mvar')},
                        'trafo3w': {'side': ('hv', 'mv', 'lv'),
                                    'meas_type': ('i_ka', 'ia_degree', 'p_mw', 'q_mvar')}}

    # Added degree result for branches
    for br_type in branch_meas_type.keys():
        for side in branch_meas_type[br_type]['side']:
            p, q, vm, va = net["res_" + br_type]["p_%s_mw" % side].values, \
                           net["res_" + br_type]["q_%s_mvar" % side].values, \
                           net["res_" + br_type]["vm_%s_pu" % side].values, \
                           net["res_" + br_type]["va_%s_degree" % side].values
            S = p + q * 1j
            V = vm * np.exp(np.deg2rad(va) * 1j)
            I = np.conj(S / V)
            net["res_" + br_type]["ia_%s_degree" % side] = np.rad2deg(np.angle(I))

    for bus_ix, bus_res in net.res_bus.iterrows():
        for meas_type in bus_meas_types.keys():
            meas_value = float(bus_res[bus_meas_types[meas_type]])
            if meas_type in ('p', 'q'):
                create_measurement(net, meas_type=meas_type, element_type='bus', element=bus_ix,
                                      value=meas_value, std_dev=1)
            else:
                create_measurement(net, meas_type=meas_type, element_type='bus', element=bus_ix,
                                      value=meas_value, std_dev=v_std_dev)

    for br_type in branch_meas_type.keys():
        if not net['res_' + br_type].empty:
            for br_ix, br_res in net['res_' + br_type].iterrows():
                for side in branch_meas_type[br_type]['side']:
                    for meas_type in branch_meas_type[br_type]['meas_type']:
                        create_measurement(net, meas_type=meas_type.split("_")[0], element_type=br_type,
                                              element=br_ix, side=side,
                                              value=br_res[meas_type.split("_")[0] + '_' +
                                                           side + '_' + meas_type.split("_")[1]], std_dev=1)
    if with_random_error:
        add_virtual_meas_error(net, v_std_dev=v_std_dev, p_std_dev=p_std_dev, q_std_dev=q_std_dev,
                               i_std_dev=i_std_dev, dg_std_dev=dg_std_dev, with_random_error=with_random_error)


def add_virtual_meas_error(net, v_std_dev=0.001, i_std_dev=0.01,
                           p_std_dev=0.03, q_std_dev=0.03, dg_std_dev=0.1, with_random_error=True):
    if net.measurement.empty:
        raise AssertionError("Measurement cannot be empty!")

    r = np.random.normal(0, 1, net.measurement.shape[0])  # random error in range from -1, 1
    p_meas_mask = net.measurement.measurement_type == "p"
    q_meas_mask = net.measurement.measurement_type == "q"
    v_meas_mask = net.measurement.measurement_type == "v"
    i_meas_mask = net.measurement.measurement_type == "i"
    dg_meas_mask = net.measurement.measurement_type.isin(("ia", "va"))

    if with_random_error:
        net.measurement.loc[p_meas_mask, 'value'] += r[p_meas_mask.values] * p_std_dev *\
            net.measurement.loc[p_meas_mask, 'value'].abs()
        net.measurement.loc[q_meas_mask, 'value'] += r[q_meas_mask.values] * q_std_dev *\
            net.measurement.loc[q_meas_mask, 'value'].abs()
        net.measurement.loc[v_meas_mask, 'value'] += r[v_meas_mask.values] * v_std_dev
        net.measurement.loc[i_meas_mask, 'value'] += r[i_meas_mask.values] * i_std_dev / 10000
        net.measurement.loc[dg_meas_mask, 'value'] += r[dg_meas_mask.values] * dg_std_dev

    net.measurement.loc[p_meas_mask, 'std_dev'] = p_std_dev
    net.measurement.loc[q_meas_mask, 'std_dev'] = q_std_dev
    net.measurement.loc[v_meas_mask, 'std_dev'] = v_std_dev
    net.measurement.loc[i_meas_mask, 'std_dev'] = i_std_dev
    net.measurement.loc[dg_meas_mask, 'std_dev'] = dg_std_dev



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

from pandapower.control.controller.characteristic_control import CharacteristicControl


class TapDependentImpedance(CharacteristicControl):
    """
    Controller that adjusts the impedance of a transformer (or multiple transformers) depending to the actual tap position and
    according to a defined characteristic.

    INPUT:
        **net** (attrdict) - Pandapower net

        **element_index** (int or list or numpy array) - ID of the transformer or multiple transfromers

        **characteristic** (object of class Characteristic) - Characteristic that describes the relationship between transformer tap
                                                                position and transformer impedance

    OPTIONAL:

        **in_service** (bool, True) - Indicates if the controller is currently in_service

        **drop_same_existing_ctrl** (bool, False) - Indicates if already existing controllers of the same type and with the same matching parameters (e.g. at same element) should be dropped
    """

    def __init__(self, net, element_index, characteristic_index, element="trafo",
                 output_variable="vk_percent", tol=1e-3, restore=True, in_service=True, order=0,
                 level=0, drop_same_existing_ctrl=False, matching_params=None, **kwargs):
        if matching_params is None:
            matching_params = {"element_index": element_index, 'output_variable': output_variable}
            super().__init__(
                net, output_element=element, output_variable=output_variable,
                output_element_index=element_index, input_element=element,
                input_variable="tap_pos", input_element_index=element_index,
                characteristic_index=characteristic_index, tol=tol, in_service=in_service,
                order=order, level=level, drop_same_existing_ctrl=drop_same_existing_ctrl,
                matching_params=matching_params, **kwargs)
        self.restore=restore
        self.initial_values = net[element].loc[element_index, output_variable].copy()

    def initialize_control(self, net):
        if self.restore:
            self.initial_values = net[self.output_element].loc[
                self.output_element_index, self.output_variable].copy()

    def finalize_control(self, net):
        if self.restore:
            net[self.output_element].loc[self.output_element_index, self.output_variable] = \
                self.initial_values



import logging
import time

import pandas as pd

from pandapower.converter.cim import cim_tools
from pandapower.converter.cim.cim2pp import build_pp_net
from pandapower.converter.cim.other_classes import Report, LogLevel, ReportCode

logger = logging.getLogger('cim.cim2pp.converter_classes.switchesCim16')

sc = cim_tools.get_pp_net_special_columns_dict()


class SwitchesCim16:
    def __init__(self, cimConverter: build_pp_net.CimConverter):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.cimConverter = cimConverter

    def convert_switches_cim16(self):
        time_start = time.time()
        self.logger.info("Start converting Breakers, Disconnectors, LoadBreakSwitches and Switches.")
        eqssh_switches = self._prepare_switches_cim16()
        self.cimConverter.copy_to_pp('switch', eqssh_switches)
        self.logger.info("Created %s switches in %ss." % (eqssh_switches.index.size, time.time() - time_start))
        self.cimConverter.report_container.add_log(Report(
            level=LogLevel.INFO, code=ReportCode.INFO_CONVERTING,
            message="Created %s switches from Breakers, Disconnectors, LoadBreakSwitches and Switches in %ss." %
                    (eqssh_switches.index.size, time.time() - time_start)))

    def _prepare_switches_cim16(self) -> pd.DataFrame:
        eqssh_switches = self.cimConverter.merge_eq_ssh_profile('Breaker', add_cim_type_column=True)
        eqssh_switches['type'] = 'CB'
        start_index_cim_net = eqssh_switches.index.size
        eqssh_switches = \
            pd.concat(
                [eqssh_switches, self.cimConverter.merge_eq_ssh_profile('Disconnector', add_cim_type_column=True)],
                ignore_index=True, sort=False)
        eqssh_switches.loc[start_index_cim_net:, 'type'] = 'DS'
        start_index_cim_net = eqssh_switches.index.size
        eqssh_switches = \
            pd.concat(
                [eqssh_switches, self.cimConverter.merge_eq_ssh_profile('LoadBreakSwitch', add_cim_type_column=True)],
                ignore_index=True, sort=False)
        eqssh_switches.loc[start_index_cim_net:, 'type'] = 'LBS'
        start_index_cim_net = eqssh_switches.index.size
        # switches needs to be the last which getting appended because of class inherit problem in jpa
        eqssh_switches = pd.concat(
            [eqssh_switches, self.cimConverter.merge_eq_ssh_profile('Switch', add_cim_type_column=True)],
            ignore_index=True, sort=False)
        eqssh_switches.loc[start_index_cim_net:, 'type'] = 'LS'
        # drop all duplicates to fix class inherit problem in jpa
        eqssh_switches = eqssh_switches.drop_duplicates(subset=['rdfId'], keep='first')
        switch_length_before_merge = eqssh_switches.index.size
        # until now eqssh_switches looks like:
        #   rdfId   name    open    ...
        #   _x01    switch1 True    ...
        #   _x02    switch2 False   ...
        # now join with the terminals
        eqssh_switches = pd.merge(eqssh_switches, self.cimConverter.bus_merge, how='left', on='rdfId')
        eqssh_switches = eqssh_switches.sort_values(by=['rdfId', 'sequenceNumber'])
        eqssh_switches = eqssh_switches.reset_index()
        # copy the columns which are needed to reduce the eqssh_switches to one line per switch
        eqssh_switches['rdfId_Terminal2'] = eqssh_switches['rdfId_Terminal'].copy()
        eqssh_switches['connected2'] = eqssh_switches['connected'].copy()
        eqssh_switches['index_bus2'] = eqssh_switches['index_bus'].copy()
        if eqssh_switches.index.size == switch_length_before_merge * 2:
            # here is where the magic happens: just remove the first value from the copied columns, reset the index
            # and replace the old column with the cut one. At least just remove the duplicates on column rdfId
            eqssh_switches.rdfId_Terminal2 = eqssh_switches.rdfId_Terminal2.iloc[1:].reset_index().rdfId_Terminal2
            eqssh_switches.connected2 = eqssh_switches.connected2.iloc[1:].reset_index().connected2
            eqssh_switches.index_bus2 = eqssh_switches.index_bus2.iloc[1:].reset_index().index_bus2
            eqssh_switches = eqssh_switches.drop_duplicates(subset=['rdfId'], keep='first')
        else:
            self.logger.error("Something went wrong at switches, seems like that terminals for connection with "
                              "connectivity nodes are missing!")
            self.cimConverter.report_container.add_log(Report(
                level=LogLevel.ERROR, code=ReportCode.ERROR_CONVERTING,
                message="Something went wrong at switches, seems like that terminals for connection with "
                        "connectivity nodes are missing!"))
            dups = eqssh_switches.pivot_table(index=['rdfId'], aggfunc='size')
            dups = dups.loc[dups != 2]
            for rdfId, count in dups.items():
                self.logger.warning("The switch with RDF ID %s has %s Terminals!" % (rdfId, count))
                self.logger.warning("The switch data: \n%s" % eqssh_switches[eqssh_switches['rdfId'] == rdfId])
                self.cimConverter.report_container.add_log(Report(
                    level=LogLevel.WARNING, code=ReportCode.WARNING_CONVERTING,
                    message="The switch with RDF ID %s has %s Terminals!" % (rdfId, count)))
            eqssh_switches = eqssh_switches[0:0]
        eqssh_switches = eqssh_switches.rename(columns={'rdfId': sc['o_id'], 'index_bus': 'bus', 'index_bus2': 'element',
                                       'rdfId_Terminal': sc['t_bus'], 'rdfId_Terminal2': sc['t_ele']})
        eqssh_switches['et'] = 'b'
        eqssh_switches['z_ohm'] = 0
        if 'inService' not in eqssh_switches.columns:
            eqssh_switches['inService'] = True
        if eqssh_switches.index.size > 0:
            eqssh_switches['closed'] = (~eqssh_switches.open & eqssh_switches.connected & eqssh_switches.connected2
                                        & eqssh_switches.inService)
        return eqssh_switches



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2023 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.


import json
import math
import os
import tempfile

from os import remove
from os.path import isfile
from collections import OrderedDict
import numpy as np
import pandas as pd

from pandapower.auxiliary import _add_ppc_options, _add_opf_options, _add_auxiliary_elements
from pandapower.build_branch import _calc_line_parameter
from pandapower.pd2ppc import _pd2ppc
from pandapower.pypower.idx_brch import ANGMIN, ANGMAX, BR_R, BR_X, BR_B, BR_G, RATE_A, RATE_B, RATE_C, \
    TAP, SHIFT, branch_cols, F_BUS, T_BUS, BR_STATUS
from pandapower.pypower.idx_bus import ZONE, VA, BASE_KV, BS, GS, BUS_I, BUS_TYPE, VMAX, VMIN, \
     VM, PD, QD
from pandapower.pypower.idx_cost import MODEL, NCOST, COST
from pandapower.pypower.idx_gen import PG, QG, GEN_BUS, VG, GEN_STATUS, QMAX, QMIN, PMIN, PMAX
from pandapower.results import init_results


# const value in branch for tnep
CONSTRUCTION_COST = 23
try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (np.int_, np.intc, np.intp, np.int8,
                            np.int16, np.int32, np.int64, np.uint8,
                            np.uint16, np.uint32, np.uint64)):
            return int(obj)
        elif isinstance(obj, (np.float64, np.float16, np.float32, np.float64)):
            return float(obj)
        elif isinstance(obj, (np.complex128, np.complex64, np.complex128)):
            return {'real': obj.real, 'imag': obj.imag}
        elif isinstance(obj, (np.ndarray,)):
            return obj.tolist()
        elif isinstance(obj, (np.bool_)):
            return bool(obj)
        elif isinstance(obj, (np.void)):
            return None
        return json.JSONEncoder.default(self, obj)


def convert_pp_to_pm(net, pm_file_path=None, correct_pm_network_data=True,
                     calculate_voltage_angles=True,
                     ac=True, silence=True, trafo_model="t", delta=1e-8, trafo3w_losses="hv",
                     check_connectivity=True, pp_to_pm_callback=None, pm_model="ACPPowerModel",
                     pm_solver="ipopt",
                     pm_mip_solver="cbc", pm_nl_solver="ipopt", opf_flow_lim="S", pm_tol=1e-8,
                     voltage_depend_loads=False, from_time_step=None, to_time_step=None, **kwargs):
    """
    Converts a pandapower net to a PowerModels.jl datastructure and saves it to a json file
    INPUT:
        **net**  - pandapower net
    OPTIONAL:
        **pm_file_path** (str, None) - Specifiy the filename, under which the .json file for
        powermodels is stored. If you want to keep the file after optimization, you should also
        set delete_buffer_file to False!

        **correct_pm_network_data** (bool, True) - checks if network data is correct.
        If not tries to correct it

        **silence** (bool, True) - Suppresses information and warning messages output by PowerModels

        **pm_model** (str, "ACPPowerModel") - The PowerModels.jl model to use

        **pm_solver** (str, "ipopt") - The "main" power models solver

        **pm_mip_solver** (str, "cbc") - The mixed integer solver (when "main" solver == juniper)

        **pm_nl_solver** (str, "ipopt") - The nonlinear solver (when "main" solver == juniper)

        **pm_time_limits** (Dict, None) - Time limits in seconds for power models interface.
        To be set as a dict like
        {"pm_time_limit": 300., "pm_nl_time_limit": 300., "pm_mip_time_limit": 300.}

        **pm_log_level** (int, 0) - solver log level in power models

        **delete_buffer_file** (Bool, True) - If True, the .json file used by powermodels will be
        deleted after optimization.

        **pp_to_pm_callback** (function, None) - callback function to add data to the PowerModels
        data structure

        **opf_flow_lim** (str, "I") - Quantity to limit for branch flow constraints, in line with
        matpower's "opf.flowlim" parameter:

            "S" - apparent power flow (limit in MVA),

            "I" - current magnitude (limit in MVA at 1 p.u. voltage)

        **pm_tol** (float, 1e-8) - default desired convergence tolerance for solver to use.

        **voltage_depend_loads** (bool, False) - consideration of voltage-dependent loads.
        If False, net.load.const_z_percent and net.load.const_i_percent are not considered,
        i.e. net.load.p_mw and net.load.q_mvar are considered as constant-power loads.

    Returns
    -------
    """

    net._options = {}

    _add_ppc_options(net, calculate_voltage_angles=calculate_voltage_angles,
                     trafo_model=trafo_model, check_connectivity=check_connectivity,
                     mode="opf", switch_rx_ratio=2, init_vm_pu="flat", init_va_degree="flat",
                     enforce_q_lims=True, recycle=dict(_is_elements=False, ppc=False, Ybus=False),
                     voltage_depend_loads=voltage_depend_loads, delta=delta,
                     trafo3w_losses=trafo3w_losses)
    _add_opf_options(net, trafo_loading='power', ac=ac, init="flat", numba=True,
                     pp_to_pm_callback=pp_to_pm_callback, pm_solver=pm_solver, pm_model=pm_model,
                     correct_pm_network_data=correct_pm_network_data, silence=silence,
                     pm_mip_solver=pm_mip_solver,
                     pm_nl_solver=pm_nl_solver, opf_flow_lim=opf_flow_lim, pm_tol=pm_tol)

    net, pm, ppc, ppci = convert_to_pm_structure(net, from_time_step=from_time_step, 
                                                 to_time_step=to_time_step)
    buffer_file = dump_pm_json(pm, pm_file_path)
    if pm_file_path is None and isfile(buffer_file):
        remove(buffer_file)
    return pm


logger = logging.getLogger(__name__)


def convert_to_pm_structure(net, opf_flow_lim="S", from_time_step=None, to_time_step=None, 
                            **kwargs):
    if net["_options"]["voltage_depend_loads"] and not (
            np.allclose(net.load.const_z_percent.values, 0) and
            np.allclose(net.load.const_i_percent.values, 0)):
        logger.error("pandapower optimal_powerflow does not support voltage depend loads.")
    net["OPF_converged"] = False
    net["converged"] = False
    _add_auxiliary_elements(net)
    init_results(net)
    ppc, ppci = _pd2ppc(net)
    ppci = build_ne_branch(net, ppci)
    net["_ppc_opf"] = ppci
    pm = ppc_to_pm(net, ppci)
    # todo: somewhere here should RATE_A be converted to 0., because only PowerModels uses 0 as no limits (pypower opf converts the zero to inf)
    pm = add_pm_options(pm, net)
    pm = add_params_to_pm(net, pm)
    if from_time_step is not None and to_time_step is not None:
        pm = add_time_series_to_pm(net, pm, from_time_step, to_time_step)
    pm = allow_multi_ext_grids(net, pm)
    net._pm = pm
    return net, pm, ppc, ppci


def dump_pm_json(pm, buffer_file=None):
    # dump pm dict to buffer_file (*.json)
    if buffer_file is None:
        # if no buffer file is provided a random file name is generated
        temp_name = next(tempfile._get_candidate_names())
        buffer_file = os.path.join(tempfile.gettempdir(), "pp_to_pm_" + temp_name + ".json")
    logger.debug("writing PowerModels data structure to %s" % buffer_file)
    with open(buffer_file, 'w') as outfile:
        json.dump(pm, outfile, indent=4, sort_keys=True, cls=NumpyEncoder)
    return buffer_file


def _pp_element_to_pm(net, pm, element, pd_bus, qd_bus, load_idx):
    bus_lookup = net._pd2ppc_lookups["bus"]

    pm_lookup = np.ones(max(net[element].index) + 1, dtype=np.int64) * -1 if len(net[element].index) \
        else np.array([], dtype=np.int64)
    for idx in net[element].index:
        if "controllable" in net[element] and net[element].at[idx, "controllable"]:
            continue

        pp_bus = net[element].at[idx, "bus"]
        pm_bus = bus_lookup[pp_bus] + 1

        scaling = net[element].at[idx, "scaling"]
        if element == "sgen":
            pd = -net[element].at[idx, "p_mw"] * scaling
            qd = -net[element].at[idx, "q_mvar"] * scaling
        else:
            pd = net[element].at[idx, "p_mw"] * scaling
            qd = net[element].at[idx, "q_mvar"] * scaling
        in_service = net[element].at[idx, "in_service"]

        pm["load"][str(load_idx)] = {"pd": pd.item(), "qd": qd.item(), "load_bus": pm_bus.item(),
                                     "status": int(in_service), "index": load_idx}
        if pm_bus not in pd_bus:
            pd_bus[pm_bus] = pd
            qd_bus[pm_bus] = qd
        else:
            pd_bus[pm_bus] += pd
            qd_bus[pm_bus] += qd

        pm_lookup[idx] = load_idx
        load_idx += 1
    return load_idx, pm_lookup


def get_branch_angles(row, correct_pm_network_data):
    angmin = row[ANGMIN].real
    angmax = row[ANGMAX].real
    # check if angles are too small for PowerModels OPF (recommendation from Carleton Coffrin
    # himself)
    if correct_pm_network_data:
        if angmin < -60.:
            logger.debug("changed voltage angle minimum of branch {}, "
                         "to -60 from {} degrees".format(int(row[0].real), angmin))
            angmin = -60.
        if angmax > 60.:
            logger.debug("changed voltage angle maximum of branch {} to 60. "
                         "from {} degrees".format(int(row[0].real), angmax))
            angmax = 60.
    # convert to rad (per unit value)
    angmin = math.radians(angmin) #/ (360/(2*np.pi))
    angmax = math.radians(angmax) #/ (360/(2*np.pi))
    return angmin, angmax


def create_pm_lookups(net, pm_lookup):
    for key, val in net._pd2ppc_lookups.items():
        if isinstance(val, dict):
            # lookup is something like "branch" with dict as val -> iterate over the subdicts
            pm_val = dict()
            for subkey, subval in val.items():
                pm_val[subkey] = tuple((v + 1 for v in subval))
        elif isinstance(val, int) or isinstance(val, np.ndarray):
            # lookup is a numpy array
            # julia starts counting at 1 instead of 0
            pm_val = val + 1
            # restore -1 for not existing elements
            pm_val[pm_val == 0] = -1
        else:
            # val not supported
            continue
        pm_lookup[key] = pm_val
    net._pd2pm_lookups = pm_lookup
    return net


def ppc_to_pm(net, ppci):
    # create power models dict. Similar to matpower case file. ne_branch is for a tnep case
    # "per_unit == True" means that the grid data in PowerModels are per-unit values. In this
    # ppc-to-pm process, the grid data schould be transformed according to baseMVA = 1.
    pm = {"gen": dict(), "branch": dict(), "bus": dict(), "dcline": dict(), "load": dict(),
          "storage": dict(),
          "ne_branch": dict(), "switch": dict(),
          "baseMVA": ppci["baseMVA"], "source_version": "2.0.0", "shunt": dict(),
          "sourcetype": "matpower", "per_unit": True, "name": net.name}
    baseMVA = ppci["baseMVA"]
    load_idx = 1
    shunt_idx = 1
    # PowerModels has a load model -> add loads and sgens to pm["load"]

    # temp dicts which hold the sum of p, q of loads + sgens
    pd_bus = dict()
    qd_bus = dict()
    load_idx, load_lookup = _pp_element_to_pm(net, pm, "load", pd_bus, qd_bus, load_idx)
    load_idx, sgen_lookup = _pp_element_to_pm(net, pm, "sgen", pd_bus, qd_bus, load_idx)
    load_idx, storage_lookup = _pp_element_to_pm(net, pm, "storage", pd_bus, qd_bus, load_idx)
    pm_lookup = {"load": load_lookup, "sgen": sgen_lookup, "storage": storage_lookup}
    net = create_pm_lookups(net, pm_lookup)

    correct_pm_network_data = net._options["correct_pm_network_data"]

    for row in ppci["bus"]:
        bus = dict()
        idx = int(row[BUS_I]) + 1
        bus["index"] = idx
        bus["bus_i"] = idx
        bus["zone"] = int(row[ZONE])
        bus["bus_type"] = int(row[BUS_TYPE])
        bus["vmax"] = row[VMAX]
        bus["vmin"] = row[VMIN]
        bus["va"] = row[VA]
        bus["vm"] = row[VM]
        bus["base_kv"] = row[BASE_KV]

        pd_value = row[PD]
        qd_value = row[QD]

        # pd and qd are the PQ values in the ppci, if they are equal to the sum in load data is
        # consistent
        if idx in pd_bus:
            pd_value -= pd_bus[idx]
            qd_value -= qd_bus[idx]
        # if not we have to add more loads wit the remaining value
        pq_mismatch = not np.allclose(pd_value, 0.) or not np.allclose(qd_value, 0.)
        if pq_mismatch:
            # This will be called if ppc PQ != sum at bus.
            logger.info("PQ mismatch. Adding another load at idx {}".format(load_idx))
            pm["load"][str(load_idx)] = {"pd": pd_value, "qd": qd_value, "load_bus": idx,
                                         "status": True, "index": load_idx}
            load_idx += 1
        # if bs or gs != 0. -> shunt element at this bus
        bs = row[BS] / baseMVA # to be validated
        gs = row[GS] / baseMVA # to be validated
        if not np.allclose(bs, 0.) or not np.allclose(gs, 0.):
            pm["shunt"][str(shunt_idx)] = {"gs": gs, "bs": bs, "shunt_bus": idx,
                                           "status": True, "index": shunt_idx}
            shunt_idx += 1
        pm["bus"][str(idx)] = bus

    n_lines = net.line.in_service.sum()
    for idx, row in enumerate(ppci["branch"], start=1):
        branch = dict()
        branch["index"] = idx
        branch["transformer"] = bool(idx > n_lines)
        branch["br_r"] = row[BR_R].real / baseMVA
        branch["br_x"] = row[BR_X].real / baseMVA
        branch["g_fr"] = row[BR_G] / 2.0 / baseMVA
        branch["g_to"] = row[BR_G] / 2.0 / baseMVA
        branch["b_fr"] = row[BR_B] / 2.0 * baseMVA
        branch["b_to"] = row[BR_B] / 2.0 * baseMVA

        if net._options["opf_flow_lim"] == "S":  # or branch["transformer"]:
            branch["rate_a"] = row[RATE_A].real if row[RATE_A] > 0 else row[RATE_B].real
            branch["rate_b"] = row[RATE_B].real
            branch["rate_c"] = row[RATE_C].real
        elif net._options["opf_flow_lim"] == "I":  # need to call _run_opf_cl from PowerModels
            f = net._pd2ppc_lookups["branch"]["line"][0]
            f = int(row[F_BUS].real)  # from bus of this line
            vr = ppci["bus"][f][BASE_KV]
            branch["c_rating_a"] = row[RATE_A].real if row[RATE_A] > 0 else row[RATE_B].real
            branch["c_rating_a"] = branch["c_rating_a"]
            branch["c_rating_b"] = row[RATE_B].real
            branch["c_rating_c"] = row[RATE_C].real
        else:
            logger.error("Branch flow limit %s not understood", net._options["opf_flow_lim"])

        branch["f_bus"] = int(row[F_BUS].real) + 1
        branch["t_bus"] = int(row[T_BUS].real) + 1
        branch["br_status"] = int(row[BR_STATUS].real)
        branch["angmin"], branch["angmax"] = get_branch_angles(row, correct_pm_network_data)
        branch["tap"] = row[TAP].real
        branch["shift"] = math.radians(row[SHIFT].real)
        pm["branch"][str(idx)] = branch

    #### create pm["gen"]
    gen_idxs_pm = [str(i+1) for i in range(len(ppci["gen"]))]
    gen_df = pd.DataFrame(index=gen_idxs_pm)
    gen_df["pg"] = ppci["gen"][:, PG]
    gen_df["qg"] = ppci["gen"][:, QG]
    gen_df["gen_bus"] = (ppci["gen"][:, GEN_BUS] + 1).astype(np.int64)
    gen_df["vg"] = ppci["gen"][:, VG]
    gen_df["qmax"] = ppci["gen"][:, QMAX]
    gen_df["gen_status"] = ppci["gen"][:, GEN_STATUS].astype(np.int64)
    gen_df["qmin"] = ppci["gen"][:, QMIN]
    gen_df["pmin"] = ppci["gen"][:, PMIN]
    gen_df["pmax"] = ppci["gen"][:, PMAX]
    gen_df["index"] = list(map(int, gen_idxs_pm))
    # add cost-parameters
    if len(ppci["gencost"]) > len(ppci["gen"]):
        logger.warning("PowerModels.jl does not consider reactive power cost - costs are ignored")
        ppci["gencost"] = ppci["gencost"][:ppci["gen"].shape[0], :]
    gen_df["startup"] = 0.0
    gen_df["shutdown"] = 0.0
    model_type = ppci["gencost"][:, MODEL].astype(np.int64)
    gen_df["model"] = model_type
    # calc ncost and cost
    ncost = np.array([0] * len(ppci["gen"]))
    cost = [[0, 0, 0] for i in gen_idxs_pm]
    ncost[model_type==1] = ppci["gencost"][:, NCOST][model_type==1]
    ncost[model_type==2] = 3
    for i in np.where(model_type==1)[0]:
        cost[i] = ppci["gencost"][i, COST:COST + ncost[i] * 2].tolist()
    for i in np.where(model_type==2)[0]:
        cost_value = ppci["gencost"][i, COST:].tolist()
        if len(cost_value) > 3:
            raise ValueError("Maximum quadratic cost function allowed")
        cost[i][-len(cost_value):] = cost_value
    gen_df["ncost"] = ncost
    gen_df["cost"] = cost
    # dataframe to dict
    pm["gen"] = gen_df.astype(object).T.to_dict()

    if "ne_branch" in ppci:
        for idx, row in enumerate(ppci["ne_branch"], start=1):
            branch = dict()
            branch["index"] = idx
            branch["transformer"] = False
            branch["br_r"] = row[BR_R].real / baseMVA
            branch["br_x"] = row[BR_X].real / baseMVA
            branch["g_fr"] = - row[BR_B].imag / 2.0 / baseMVA
            branch["g_to"] = - row[BR_B].imag / 2.0 / baseMVA
            branch["b_fr"] = row[BR_B].real / 2.0 * baseMVA
            branch["b_to"] = row[BR_B].real / 2.0 * baseMVA

            if net._options["opf_flow_lim"] == "S":  # --> Rate_a is always needed for the TNEP problem, right?
                branch["rate_a"] = row[RATE_A].real if row[RATE_A] > 0 else row[RATE_B].real
                branch["rate_b"] = row[RATE_B].real
                branch["rate_c"] = row[RATE_C].real
            elif net._options["opf_flow_lim"] == "I":
                f, t = net._pd2ppc_lookups["branch"]["line"]
                f = int(row[F_BUS].real)  # from bus of this line
                vr = ppci["bus"][f][BASE_KV]
                row[RATE_A] = row[RATE_A] / (vr * np.sqrt(3))

                branch["c_rating_a"] = row[RATE_A].real if row[RATE_A] > 0 else row[RATE_B].real
                branch["c_rating_b"] = row[RATE_B].real
                branch["c_rating_c"] = row[RATE_C].real

            branch["f_bus"] = int(row[F_BUS].real) + 1
            branch["t_bus"] = int(row[T_BUS].real) + 1
            branch["br_status"] = int(row[BR_STATUS].real)
            branch["angmin"], branch["angmax"] = get_branch_angles(row, correct_pm_network_data)
            branch["tap"] = row[TAP].real
            branch["shift"] = math.radians(row[SHIFT].real)
            branch["construction_cost"] = row[CONSTRUCTION_COST].real
            pm["ne_branch"][str(idx)] = branch

    return pm


def add_pm_options(pm, net):
    # read values from net_options if present else use default values
    pm["pm_solver"] = net._options["pm_solver"] if "pm_solver" in net._options else "ipopt"
    pm["pm_mip_solver"] = net._options["pm_mip_solver"] if "pm_mip_solver" in net._options else "cbc"
    pm["pm_nl_solver"] = net._options["pm_nl_solver"] if "pm_nl_solver" in net._options else "ipopt"
    pm["pm_model"] = net._options["pm_model"] if "pm_model" in net._options else "DCPPowerModel"
    pm["pm_log_level"] = net._options["pm_log_level"] if "pm_log_level" in net._options else 0
    pm["pm_tol"] = net._options["pm_tol"] if "pm_tol" in net._options else 1e-8
    if "pm_time_limits" in net._options and isinstance(net._options["pm_time_limits"], dict):
        # write time limits to power models data structure
        for key, val in net._options["pm_time_limits"].items():
            pm[key] = val
    else:
        pm["pm_time_limit"], pm["pm_nl_time_limit"], pm["pm_mip_time_limit"] = \
            np.inf, np.inf, np.inf
    pm["correct_pm_network_data"] = net._options["correct_pm_network_data"]
    pm["silence"] = net._options["silence"]
    pm["ac"] = net._options["ac"]
    return pm


def build_ne_branch(net, ppc):
    # this is only used by pm tnep
    if "ne_line" in net:
        length = len(net["ne_line"])
        ppc["ne_branch"] = np.zeros(shape=(length, branch_cols + 1), dtype=np.complex128)
        ppc["ne_branch"][:, :13] = np.array([0, 0, 0, 0, 0, 250, 250, 250, 1, 0, 1, -60, 60])
        # create branch array ne_branch like the common branch array in the ppc
        net._pd2ppc_lookups["ne_branch"] = dict()
        net._pd2ppc_lookups["ne_branch"]["ne_line"] = (0, length)
        _calc_line_parameter(net, ppc, "ne_line", "ne_branch")
        ppc["ne_branch"][:, CONSTRUCTION_COST] = net["ne_line"].loc[:, "construction_cost"].values
    return ppc


def init_ne_line(net, new_line_index, construction_costs=None):
    """
    init function for new line dataframe, which specifies the possible new lines being built by power models tnep opt
    Parameters
    ----------
    net - pp net
    new_line_index (list) - indices of new lines. These are copied to the new dataframe net["ne_line"] from net["line"]
    construction_costs (list, 0.) - costs of newly constructed lines
    Returns
    -------
    """
    # init dataframe
    net["ne_line"] = net["line"].loc[new_line_index, :]
    # add costs, if None -> init with zeros
    construction_costs = np.zeros(len(new_line_index)) if construction_costs is None else \
        construction_costs
    net["ne_line"].loc[new_line_index, "construction_cost"] = construction_costs
    # set in service, but only in ne line dataframe
    net["ne_line"].loc[new_line_index, "in_service"] = True
    # init res_ne_line to save built status afterwards
    net["res_ne_line"] = pd.DataFrame(data=0, index=new_line_index, columns=["built"], dtype=np.int64)


def add_params_to_pm(net, pm):
    # add user defined parameters to pm
    pd_idxs_br = []
    pm_idxs_br = []
    br_elms = ["line", "trafo"]
    pm["user_defined_params"] = dict()
    for elm in ["bus", "line", "gen", "load", "trafo", "sgen"]:
        param_cols = [col for col in net[elm].columns if 'pm_param' in col]
        if not param_cols:
            continue
        params = [param_col.split("/")[-1] for param_col in param_cols]
        if "side" in params:
            params.remove("side")
            params.insert(len(params), "side")
            param_cols.remove("pm_param/side")
            param_cols.insert(len(params), "pm_param/side")
            
        br_param = list(set(params) - {'side'})
        for param, param_col in zip(params, param_cols):
            pd_idxs = net[elm].index[net[elm][param_col].notna()].tolist()
            target_values = net[elm][param_col][pd_idxs].values.tolist()
            if elm in br_elms and param in br_param:
                pd_idxs_br += net[elm].index[net[elm][param_col].notna()].tolist()
                target_values = net[elm][param_col][pd_idxs_br].values.tolist()
            if elm in ["line", "trafo"]:
                start, end = net._pd2pm_lookups["branch"][elm]
                pd_pos = [net[elm].index.tolist().index(p) for p in pd_idxs_br]
                pm_idxs = [int(v) + start for v in pd_pos]
            elif elm == "sgen":
                pm_idxs = [int(v) for v in net._pd2pm_lookups[elm+"_controllable"][pd_idxs]]
                elm = "gen"
            else:
                pm_idxs = [int(v) for v in net._pd2pm_lookups[elm][pd_idxs]]
            df = pd.DataFrame(index=pm_idxs) if elm not in ["line", "trafo"] else pd.DataFrame(index=pm_idxs_br)   
            df["element_index"] = pm_idxs
            df["element_pp_index"] = pd_idxs if elm not in ["line", "trafo"] else pd_idxs_br
            df["value"] = target_values
            df["element"] = elm
            pm["user_defined_params"][param] = df.to_dict(into=OrderedDict, orient="index")

        if elm in ["line", "trafo"]:
            for bp in br_param:
                for k in pm["user_defined_params"]["side"].keys():
                    side = pm["user_defined_params"]["side"][k]["value"]
                    side_bus_f = side + "_bus"
                    if elm == "line":
                        side_bus_t = "from_bus" if side == "to" else "to_bus"
                    if elm == "trafo":
                        side_bus_t = "hv_bus" if side == "lv" else "lv_bus"
                    pd_idx = pm["user_defined_params"]["side"][k]["element_pp_index"]
                    ppcidx = net._pd2pm_lookups["branch"][elm][0]-1+pd_idx   
                    
                    if side in ["from", "hv"]:
                        ppcrow_f = 0
                        ppcrow_t = 1
                    else:
                        ppcrow_f = 1
                        ppcrow_t = 0
                        assert side in ["to", "lv"]

                    pm["user_defined_params"][bp][k]["f_bus"] = \
                        int(net._ppc_opf["branch"][ppcidx, ppcrow_f].real) + 1
                    pm["user_defined_params"][bp][k]["t_bus"] = \
                        int(net._ppc_opf["branch"][ppcidx, ppcrow_t].real) + 1
                    # pm["user_defined_params"][bp][k]["f_bus"] = \
                    #     net._pd2pm_lookups["bus"][net[elm][side_bus_f][pd_idx]]
                    # pm["user_defined_params"][bp][k]["t_bus"] = \
                    #     net._pd2pm_lookups["bus"][net[elm][side_bus_t][pd_idx]]

    # add controllable sgen:
    dic = {}
    if "user_defined_params" in pm.keys():
        for elm in ["gen", "sgen_controllable"]:
            if elm in net._pd2pm_lookups.keys():
                pm_idxs = net._pd2pm_lookups[elm] 
                for k in pm_idxs[pm_idxs!=-1]:
                    dic[str(k)] = k
        if dic != {}:
            pm["user_defined_params"]["gen_and_controllable_sgen"] = dic
    
    # add objective factors for multi optimization
    if "obj_factors" in net.keys():
        assert type(net.obj_factors) == list
        assert sum(net.obj_factors) <= 1
        dic = {}
        for i, k in enumerate(net.obj_factors):
            dic["fac_"+str(i+1)] = k        
        pm["user_defined_params"]["obj_factors"] = dic

    return pm


def add_time_series_to_pm(net, pm, from_time_step, to_time_step):
    from pandapower.control import ConstControl
    if from_time_step is None or to_time_step is None:
        raise ValueError("please define 'from_time_step' " +
                         "and 'to_time_step' to call time-series optimizaiton ")
    tp_list = list(range(from_time_step, to_time_step))
    if len(net.controller):
        load_dict, gen_dict = {}, {}
        pm["time_series"] = {"load": load_dict, "gen": gen_dict,
                             "from_time_step": from_time_step+1, 
                             "to_time_step": to_time_step+1} 
        for idx, content in net.controller.iterrows():
            if not type(content["object"]) == ConstControl:
                continue
            else:
                element = content["object"].__dict__["matching_params"]["element"]
                variable = content["object"].__dict__["matching_params"]["variable"]
                elm_idxs = content["object"].__dict__["matching_params"]["element_index"]
                df = content["object"].data_source.df
                for pd_ei in elm_idxs:
                    if element == "sgen" and "controllable" in net[element].columns and net[element].controllable[pd_ei]:
                        pm_ei = net._pd2pm_lookups[element+"_controllable"][pd_ei]
                        pm_elm = "gen"
                    else:
                        pm_ei = net._pd2pm_lookups[element][pd_ei]
                        pm_elm = "load"
                    if str(pm_ei) not in list(pm["time_series"][pm_elm].keys()):
                        pm["time_series"][pm_elm][str(pm_ei)] = {}
                    target_ts = df[pd_ei][from_time_step:to_time_step].values
                    pm["time_series"][pm_elm][str(pm_ei)][variable] = \
                       {str(tp_list[m]):n for m, n in enumerate(list(-target_ts))} if (element!="load" and pm_elm not in element) \
                            else {str(tp_list[m]):n for m, n in enumerate(list(target_ts))} 
    return pm


def allow_multi_ext_grids(net, pm, ext_grids=None):
    ext_grids = net.ext_grid.index.tolist() if ext_grids is None else ext_grids
    target_pp_buses = net.ext_grid.bus[ext_grids]
    target_pm_buses = net._pd2pm_lookups["bus"][target_pp_buses]
    for b in target_pm_buses:
        pm["bus"][str(b)]["bus_type"] = 3
    return pm



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.



try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)

# separator between log messages
log_message_sep = ("\n --------\n")


def diagnostic_report(net, diag_results, diag_errors, diag_params, compact_report, warnings_only):
    diag_report = DiagnosticReports(net, diag_results, diag_errors, diag_params, compact_report)
    if warnings_only:
        logger.setLevel(logging.WARNING)
    else:
        logger.setLevel(logging.INFO)

    report_methods = {
        "missing_bus_indices": diag_report.report_missing_bus_indices,
        "disconnected_elements": diag_report.report_disconnected_elements,
        "different_voltage_levels_connected": diag_report.report_different_voltage_levels_connected,
        "impedance_values_close_to_zero": diag_report.report_impedance_values_close_to_zero,
        "nominal_voltages_dont_match": diag_report.report_nominal_voltages_dont_match,
        "invalid_values": diag_report.report_invalid_values,
        "overload": diag_report.report_overload,
        "multiple_voltage_controlling_elements_per_bus": diag_report.report_multiple_voltage_controlling_elements_per_bus,
        "wrong_switch_configuration": diag_report.report_wrong_switch_configuration,
        "no_ext_grid": diag_report.report_no_ext_grid,
        "wrong_reference_system": diag_report.report_wrong_reference_system,
        "deviation_from_std_type": diag_report.report_deviation_from_std_type,
        "numba_comparison": diag_report.report_numba_comparison,
        "parallel_switches": diag_report.report_parallel_switches
    }

    logger.warning("\n\n_____________ PANDAPOWER DIAGNOSTIC TOOL _____________ \n")
    for key in report_methods:
        if (key in diag_results) or not warnings_only:
            report_methods[key]()
            logger.warning(log_message_sep)

    logger.warning("_____________ END OF PANDAPOWER DIAGNOSTIC _____________ ")


class DiagnosticReports:
    def __init__(self, net, diag_results, diag_errors, diag_params, compact_report):
        self.net = net
        self.diag_results = diag_results
        self.diag_errors = diag_errors
        self.diag_params = diag_params
        self.compact_report = compact_report

    def report_disconnected_elements(self):
        if "disconnected_elements" in self.diag_results:
            # message header
            if self.compact_report:
                logger.warning("disconnected_elements:")
            else:
                logger.warning("Checking for elements without a connection to an external grid...")
            logger.warning("")

            # message body
            diag_result = self.diag_results["disconnected_elements"]
            element_counter = 0
            for disc_section in diag_result:
                if self.compact_report:
                    logger.warning("disonnected_section: %s" % (disc_section))

                else:
                    logger.warning("Disconnected section found,"
                                   " consisting of the following elements:")
                    for key in disc_section:
                        element_counter += len(disc_section[key])
                        logger.warning("%s: %s" % (key, disc_section[key]))

            # message summary
            if not self.compact_report:
                logger.warning("")
                logger.warning("SUMMARY: %s disconnected element(s) found." % (element_counter))
        elif "disconnected_elements" in self.diag_errors:
            logger.warning("Check for disconnected elements failed due to the following error:")
            logger.warning(self.diag_errors["disconnected_elements"])
        else:
            logger.info("PASSED: No problematic switches found")

    def report_different_voltage_levels_connected(self):
        from pandapower.toolbox import get_connected_buses_at_element

        if "different_voltage_levels_connected" in self.diag_results:

            # message header
            if self.compact_report:
                logger.warning("different_voltage_levels_connected:")

            else:
                logger.warning("Checking for connections of different voltage levels...")
            logger.warning("")

            # message body
            diag_result = self.diag_results["different_voltage_levels_connected"]
            element_counter = 0
            for key in diag_result:
                element_counter += len(diag_result[key])
                if key == "lines":
                    element_type = "line"
                elif key == "switches":
                    element_type = "switch"
                if self.compact_report:
                    logger.warning("%s:" % (key))
                for element in diag_result[key]:
                    buses = list(get_connected_buses_at_element(self.net, element,
                                                                key[0]))
                    if self.compact_report:
                        logger.warning("%s %s: buses %s" % (element_type, element, buses))
                    else:
                        logger.warning("%s %s connects bus %s: %s (vn_kv = %s) "
                                       "and bus %s: %s (vn_kv = %s)"
                                       % (element_type, element, buses[0],
                                          self.net.bus.name.at[buses[0]],
                                          self.net.bus.vn_kv.at[buses[0]],
                                          buses[1], self.net.bus.name.at[buses[1]],
                                          self.net.bus.vn_kv.at[buses[1]]))
                        # message summary
            if not self.compact_report:
                logger.warning("")
                logger.warning("SUMMARY: %s element(s) that connect different voltage "
                               "levels found." % (element_counter))
        elif "different_voltage_levels_connected" in self.diag_errors:
            logger.warning("Check for connection of different voltage levels failed due to the following error:")
            logger.warning(self.diag_errors["different_voltage_levels_connected"])
        else:
            logger.info("PASSED: No connection of different voltage levels found")


    def report_impedance_values_close_to_zero(self):

        if "impedance_values_close_to_zero" in self.diag_results:

            # message header
            if self.compact_report:
                logger.warning("impedance_values_close_to_zero:")

            else:
                logger.warning("Checking for impedance values close to zero...")
            logger.warning("")

            # message body
            diag_result = self.diag_results["impedance_values_close_to_zero"][0]
            element_counter = 0
            for key in diag_result:
                element_counter += len(diag_result[key])
                for element in diag_result[key]:
                    if key in ("line", "line_dc"):
                        min_r_type = "r_ohm"
                        min_x_type = "x_ohm"
                    elif key == "xward":
                        min_r_type = "r_ohm"
                        min_x_type = "x_ohm"
                    elif key == "impedance":
                        min_r_type = "r_pu"
                        min_x_type = "x_pu"
                    elif key == "vsc":
                        min_r_type = "r_dc_ohm"
                        min_x_type = "x_ohm"

                    logger.warning("%s %s: %s <= %s or %s <= %s"
                                   % (key, element, min_r_type, self.diag_params["min_"+min_r_type],
                                      min_x_type, self.diag_params["min_"+min_x_type]))

            if len(self.diag_results["impedance_values_close_to_zero"]) > 1:
                switch_replacement = self.diag_results["impedance_values_close_to_zero"][1]
                if switch_replacement["loadflow_converges_with_switch_replacement"]:
                    logger.warning("Switch replacement successful: Power flow converges after "
                                   "replacing implausible elements with switches.")
                else:
                    logger.warning("Power flow still does not converge after replacing implausible "
                                   "elements with switches.")

            # message summary
            if not self.compact_report:
                logger.warning("")
                logger.warning("SUMMARY: %s element(s) with impedance values close to zero found."
                               % (element_counter))
        elif "impedance_values_close_to_zero" in self.diag_errors:
            logger.warning("Check for elements with impedance values close to zero failed due "
                           "to the following error:")
            logger.warning(self.diag_errors["impedance_values_close_to_zero"])
        else:
            logger.info("PASSED: No elements with impedance values close to zero found...")


    def report_nominal_voltages_dont_match(self):

        if "nominal_voltages_dont_match" in self.diag_results:

            # message header
            if self.compact_report:
                logger.warning("nominal_voltages_dont_match:")

            else:
                logger.warning("Checking for components with deviating nominal voltages...")
            logger.warning("")

            # message body
            diag_result = self.diag_results["nominal_voltages_dont_match"]
            nom_voltage_tolerance = self.diag_params["nom_voltage_tolerance"]
            element_counter = 0
            for element in diag_result:
                if self.compact_report:
                    logger.warning("%s:" % (element))
                for key in diag_result[element]:
                    element_counter += len(diag_result[element][key])
                    if element == "trafo":
                        if self.compact_report:
                            logger.warning("%s: %s" % (key, diag_result[element][key]))
                        else:
                            if key == "hv_lv_swapped":
                                logger.warning("Trafo(s) %s: hv and lv connectors seem to "
                                               "be swapped" % (diag_result[element][key]))
                            elif key == "hv_bus":
                                for trafo in diag_result[element][key]:
                                    logger.warning("Trafo %s: Nominal voltage on hv_side "
                                                   "(%s kV) and voltage_level of hv_bus "
                                                   "(bus %s with voltage_level %s kV) "
                                                   "deviate more than +/- %s percent."
                                                   % (trafo, self.net.trafo.vn_hv_kv.at[trafo],
                                                      self.net.trafo.hv_bus.at[trafo],
                                                      self.net.bus.vn_kv.at[self.net.trafo.hv_bus.at[trafo]],
                                                      nom_voltage_tolerance * 100))
                            elif key == "lv_bus":
                                for trafo in diag_result[element][key]:
                                    logger.warning("Trafo %s: Nominal voltage on lv_side "
                                                   "(%s kV) and voltage_level of lv_bus "
                                                   "(bus %s with voltage_level %s kV) "
                                                   "deviate more than +/- %s percent."
                                                   % (trafo, self.net.trafo.vn_lv_kv.at[trafo],
                                                      self.net.trafo.lv_bus.at[trafo],
                                                      self.net.bus.vn_kv.at[self.net.trafo.lv_bus.at[trafo]],
                                                      nom_voltage_tolerance * 100))
                    if element == "trafo3w":
                        if self.compact_report:
                            logger.warning("%s: %s" % (key, diag_result[element][key]))
                        else:
                            if key == "connectors_swapped_3w":
                                logger.warning("Trafo3w %s: connectors seem to "
                                               "be swapped" % (diag_result[element][key]))
                            elif key == "hv_bus":
                                for trafo3w in diag_result[element][key]:
                                    logger.warning("Trafo3w %s: Nominal voltage on hv_side "
                                                   "(%s kV) and voltage_level of hv_bus "
                                                   "(bus %s with voltage_level %s kV) "
                                                   "deviate more than +/- %s percent."
                                                   % (trafo3w, self.net.trafo3w.vn_hv_kv.at[trafo3w],
                                                      self.net.trafo3w.hv_bus.at[trafo3w],
                                                      self.net.bus.vn_kv.at[self.net.trafo3w.hv_bus.at[trafo3w]],
                                                      nom_voltage_tolerance * 100))
                            elif key == "mv_bus":
                                for trafo3w in diag_result[element][key]:
                                    logger.warning("Trafo3w %s: Nominal voltage on mv_side "
                                                   "(%s kV) and voltage_level of mv_bus "
                                                   "(bus %s with voltage_level %s kV) "
                                                   "deviate more than +/- %s percent."
                                                   % (trafo3w, self.net.trafo3w.vn_mv_kv.at[trafo3w],
                                                      self.net.trafo3w.mv_bus.at[trafo3w],
                                                      self.net.bus.vn_kv.at[self.net.trafo3w.mv_bus.at[trafo3w]],
                                                      nom_voltage_tolerance * 100))
                            elif key == "lv_bus":
                                for trafo3w in diag_result[element][key]:
                                    logger.warning("Trafo3w %s: Nominal voltage on lv_side "
                                                   "(%s kV) and voltage_level of lv_bus "
                                                   "(bus %s with voltage_level %s kV) "
                                                   "deviate more than +/- %s percent."
                                                   % (trafo3w, self.net.trafo3w.vn_lv_kv.at[trafo3w],
                                                      self.net.trafo3w.lv_bus.at[trafo3w],
                                                      self.net.bus.vn_kv.at[self.net.trafo3w.lv_bus.at[trafo3w]],
                                                      nom_voltage_tolerance * 100))

                                    # message summary
            if not self.compact_report:
                logger.warning("")
                logger.warning("SUMMARY: %s component(s) with deviating nominal voltages found"
                               % (element_counter))
        elif "nominal_voltages_dont_match" in self.diag_errors:
            logger.warning("Check for components with deviating nominal voltages failed due "
                           "to the following error:")
            logger.warning(self.diag_errors["nominal_voltages_dont_match"])
        else:
            logger.info("PASSED: No components with deviating nominal voltages found")

    def report_invalid_values(self):

        if "invalid_values" in self.diag_results:

            # message header
            if self.compact_report:
                logger.warning("invalid_values:")
            else:
                logger.warning("Checking for invalid_values...")
            logger.warning("")

        # message body
        if "invalid_values" in self.diag_results:
            diag_result = self.diag_results["invalid_values"]
            element_counter = 0
            for element_type in diag_result:
                element_counter += len(diag_result[element_type])
                logger.warning("%s:" % (element_type))
                for inv_value in diag_result[element_type]:
                    if self.compact_report:
                        logger.warning("%s %s: '%s' = %s (restriction: %s)"
                                       % (element_type, inv_value[0], inv_value[1], inv_value[2],
                                          inv_value[3]))
                    else:
                        logger.warning("Invalid value found: '%s %s' with attribute '%s' = %s "
                                       "(data type: %s). Valid input needs to be %s."
                                       % (element_type, inv_value[0], inv_value[1], inv_value[2],
                                          type(inv_value[2]), inv_value[3]))

                        # message summary
            if not self.compact_report:
                logger.warning("")
                logger.warning("SUMMARY: %s invalid values found." % element_counter)

        elif "invalid_values" in self.diag_errors:
            logger.warning("Check for invalid values failed due to the following error:")
            logger.warning(self.diag_errors["invalid_values"])
        else:
            logger.info("PASSED: No invalid values found")


    def report_overload(self):

        if "overload" in self.diag_results:
            # message header
            if self.compact_report:
                logger.warning("overload:")
            else:
                logger.warning("Checking for overload...")
            logger.warning("")

            # message body
            diag_result = self.diag_results["overload"]
            overload_scaling_factor = self.diag_params["overload_scaling_factor"]
            if not diag_result["load"] and not diag_result["generation"]:
                logger.warning("Overload check failed: Power flow still does not "
                               "converge with load and generation scaled down to %s percent."
                               % (overload_scaling_factor * 100))
            elif (diag_result["load"] and diag_result["generation"]):
                logger.warning("overload found: Power flow converges "
                               "with load and generation scaled down to %s percent."
                               % (overload_scaling_factor * 100))
            else:
                if diag_result["load"]:
                    logger.warning("overload found: Power flow converges "
                                   "with load scaled down to %s percent."
                                   % (overload_scaling_factor * 100))
                elif diag_result["generation"]:
                    logger.warning("overload found: Power flow converges "
                                   "with generation scaled down to %s percent."
                                   % (overload_scaling_factor * 100))
        # message summary
        elif "overload" in self.diag_errors:
            logger.warning("Check for overload failed due to the following error:")
            logger.warning(self.diag_errors["overload"])
        else:
            logger.info("PASSED: Power flow converges. No overload found.")


    def report_wrong_switch_configuration(self):

        if "wrong_switch_configuration" in self.diag_results:

            # message header
            if self.compact_report:
                logger.warning("wrong_switch_configuration:")
            else:
                logger.warning("Checking switch configuration...")
            logger.warning("")

            # message body
            diag_result = self.diag_results["wrong_switch_configuration"]
            if diag_result:
                logger.warning("Possibly wrong switch configuration found: power flow "
                               "converges with all switches closed.")
            else:
                logger.warning("Power flow still does not converge with all switches closed.")

        # message summary
        elif "wrong_switch_configuration" in self.diag_errors:
            logger.warning("Check for wrong switch configuration failed due to the following error:")
            logger.warning(self.diag_errors["wrong_switch_configuration"])
        else:
            logger.info("PASSED: Power flow converges. Switch configuration seems ok.")


    def report_no_ext_grid(self):

        if "no_ext_grid" in self.diag_results:
            # message header
            if self.compact_report:
                logger.warning("no_external_grid:")
            else:
                logger.warning("Checking if there is at least one external grid...")
            logger.warning("")

            # message body
            diag_result = self.diag_results["no_ext_grid"]
            if diag_result is True:
                logger.warning("No ext_grid found. There has to be at least one ext_grid!")

        # message summary
        elif "no_ext_grid" in self.diag_errors:
            logger.warning("Check for external grid failed due to the following error:")
            logger.warning(self.diag_errors["no_ext_grid"])
        else:
            logger.info("PASSED: External grid found.")


    def report_multiple_voltage_controlling_elements_per_bus(self):

        if "multiple_voltage_controlling_elements_per_bus" in self.diag_results:
            # message header
            if self.compact_report:
                logger.warning("multiple_voltage_controlling_elements_per_bus:")
            else:
                logger.warning("Checking for multiple gens and/or external grids per bus...")
            logger.warning("")

            # message body
            diag_result = self.diag_results["multiple_voltage_controlling_elements_per_bus"]
            element_counter = 0
            for feeder_type in diag_result:
                element_counter += len(diag_result[feeder_type])
                if self.compact_report:
                    logger.warning("%s: %s" % (feeder_type, diag_result[feeder_type]))

                else:
                    for bus in diag_result[feeder_type]:
                        if feeder_type == "buses_with_mult_ext_grids":
                            logger.warning("External grids %s are connected to bus %s. Only one "
                                           "external grid per bus is allowed."
                                           % (list(self.net.ext_grid[self.net.ext_grid.bus
                                                                     == bus].index), bus))
                        elif feeder_type == "buses_with_gens_and_ext_grids":
                            logger.warning("Generator(s) %s and external grid(s) %s are connected "
                                           "to bus %s. Only one generator OR one external grid "
                                           "per bus is allowed."
                                           % (list(self.net.gen[self.net.gen.bus == bus].index),
                                              list(self.net.ext_grid[self.net.ext_grid.bus
                                                                     == bus].index), bus))

                            # message summary
            if not self.compact_report:
                logger.warning("")
                logger.warning("SUMMARY: %s bus(ses) with multiple gens and/or ext_grids "
                               "found." % (element_counter))
        elif "multiple_voltage_controlling_elements_per_bus" in self.diag_errors:
            logger.warning("Check for multiple voltage controlling elements per bus failed due "
                           "to the following error:")
            logger.warning(self.diag_errors["multiple_voltage_controlling_elements_per_bus"])
        else:
            logger.info("PASSED: No buses with multiple gens and/or ext_grids found.")


    def report_wrong_reference_system(self):

        if "wrong_reference_system" in self.diag_results:

            # message header
            if self.compact_report:
                logger.warning("wrong_reference_system:")
            else:
                logger.warning("Checking for usage of wrong reference system...")
            logger.warning("")

            # message body
            diag_result = self.diag_results["wrong_reference_system"]
            for element_type in diag_result:
                if element_type == "loads":
                    if self.compact_report:
                        logger.warning("loads %s: wrong reference system."
                                       % (diag_result[element_type]))
                    else:
                        for load in diag_result[element_type]:
                            logger.warning("Found load %s: '%s' with p_mw = %s. In load reference "
                                           "system p_mw should be positive."
                                           % (load, self.net.load.name.at[load],
                                              self.net.load.p_mw.at[load]))

                elif element_type == "gens":
                    if self.compact_report:
                        logger.warning("gens %s: wrong reference system."
                                       % (diag_result[element_type]))
                    else:
                        for gen in diag_result[element_type]:
                            logger.warning("Found gen %s: '%s' with p_mw = %s. In load reference "
                                           "system p_mw should be negative."
                                           % (gen, self.net.gen.name.at[gen], self.net.gen.p_mw.at[gen]))

                elif element_type == "sgens":
                    if self.compact_report:
                        logger.warning("sgens %s: wrong reference system."
                                       % (diag_result[element_type]))
                    else:
                        for sgen in diag_result[element_type]:
                            logger.warning("Found sgen %s: '%s' with p_mw = %s. In load reference "
                                           "system p_mw should be negative."
                                           % (sgen, self.net.sgen.name.at[sgen], self.net.sgen.p_mw.at[sgen]))

                            # message summary
            if not self.compact_report:
                logger.warning("")
                if 'loads' in diag_result:
                    logger.warning("SUMMARY: Found %s load(s) with negative p_mw. In load "
                                   "reference system, p_mw should be positive. If the intention "
                                   "was to model a constant generation, please use an sgen instead."
                                   % (len(diag_result['loads'])))
                if 'gens' in diag_result:
                    logger.warning("SUMMARY: Found %s gen(s) with positive p_mw. In load "
                                   "reference system, p_mw should be negative. If the intention "
                                   "was to model a load, please use a load instead."
                                   % (len(diag_result['gens'])))
                if 'sgens' in diag_result:
                    logger.warning("SUMMARY: Found %s sgen(s) with positive p_mw. In load "
                                   "reference system, p_mw should be negative. If the intention "
                                   "was to model a load, please use a load instead."
                                   % (len(diag_result['sgens'])))

        elif "wrong_reference_system" in self.diag_errors:
            logger.warning("Check for wrong reference system failed due to the following error:")
            logger.warning(self.diag_errors["wrong_reference_system"])
        else:
            logger.info("PASSED: correct reference system")


    def report_deviation_from_std_type(self):

        if "deviation_from_std_type" in self.diag_results:
            # message header
            if self.compact_report:
                logger.warning("deviation_from_std_type:")
            else:
                logger.warning("Checking for deviation from std type...")
            logger.warning("")

            # message body
            diag_result = self.diag_results["deviation_from_std_type"]
            element_counter = 0
            for et in diag_result:
                for eid in diag_result[et]:
                    element_counter += 1
                    values = diag_result[et][eid]
                    if values['std_type_in_lib']:
                        logger.warning("%s %s: %s = %s, std_type_value = %s"
                                       % (et, eid, values['param'], values['e_value'],
                                          values['std_type_value']))
                    else:
                        logger.warning("%s %s: No valid std_type or std_type not in net.std_types"
                                       % (et, eid))

                        # message summary
            if not self.compact_report:
                logger.warning("")
                logger.warning("SUMMARY: %s elements with deviations from std_type found."
                               % (element_counter))

        elif "deviation_from_std_type" in self.diag_errors:
            logger.warning("Check for deviation from std_type failed due to the following error:")
            logger.warning(self.diag_errors["deviation_from_std_type"])
        else:
            logger.info("PASSED: No elements with deviations from std_type found.")

    def report_numba_comparison(self):

        if "numba_comparison" in self.diag_results:
            # message header
            if self.compact_report:
                logger.warning("numba_comparison:")
            else:
                logger.warning("Checking for deviations between numba = True vs. False...")
            logger.warning("")

            # message body
            diag_result = self.diag_results["numba_comparison"]
            element_counter = 0
            for element_type in diag_result:
                for res_type in diag_result[element_type]:
                    if self.compact_report:
                        logger.warning("%s.%s absolute deviations:\n%s"
                                       % (element_type, res_type,
                                          diag_result[element_type][res_type]))
                    else:
                        for idx in diag_result[element_type][res_type].index:
                            element_counter += 1
                            dev = diag_result[element_type][res_type].loc[idx]
                            logger.warning("%s.%s at index %s: absolute deviation = %s"
                                           % (element_type, res_type, idx, dev))

                                # message summary
                if not self.compact_report:
                    logger.warning("")
                    logger.warning("SUMMARY: %s results with deviations between numba = True vs. \
                                    False found." % (element_counter))

        elif "numba_comparison" in self.diag_errors:
            logger.warning("numba_comparison failed due to the following error:")
            logger.warning(self.diag_errors["numba_comparison"])
        else:
            logger.info("PASSED: No results with deviations between numba = True vs. False found.")


    def report_parallel_switches(self):

        if "parallel_switches" in self.diag_results:
            # message header
            if self.compact_report:
                logger.warning("parallel_switches:")
            else:
                logger.warning("Checking for parallel switches...")
            logger.warning("")

            # message body
            diag_result = self.diag_results["parallel_switches"]
            for switch_tuple in diag_result:
                logger.warning("switches %s are parallel." % switch_tuple)

                # message summary
            if not self.compact_report:
                logger.warning("")
                logger.warning("SUMMARY: %s occurences of parallel switches found."
                               % len(diag_result))
        elif "parallel_switches" in self.diag_errors:
            logger.warning("Check for parallel_switches failed due to the following error:")
            logger.warning(self.diag_errors["parallel_switches"])
        else:
            logger.info("PASSED: No parallel switches found.")


    def report_missing_bus_indices(self):
        # message header
        if self.compact_report:
            logger.info("missing_bus_indices:")
        else:
            logger.info("Checking for missing bus indices...")
        logger.info("")
        if "missing_bus_indices" in self.diag_results:
            # message body
            diag_result = self.diag_results["missing_bus_indices"]
            element_counter = 0
            for element_type in diag_result:
                for element in diag_result[element_type]:
                    element_counter += 1
                    logger.warning("%s %s: %s (%s) not in net.bus.index" % (element_type,
                                                                            element[0], element[1], element[2]))

                    # message summary
            if not self.compact_report:
                logger.warning("")
                logger.warning("SUMMARY: %s missing bus indices found." % element_counter)

        elif "missing_bus_indices" in self.diag_errors:
            logger.warning("Check for missing bus indices failed due to the following error:")
            logger.warning(self.diag_errors["missing_bus_indices"])
        else:
            logger.info("PASSED: No missing bus indices found.")



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2025 by University of Kassel and Fraunhofer Institute for Energy Economics
# and Energy System Technology (IEE), Kassel. All rights reserved.

import numpy as np

from pandapower.toolbox.power_factor import cosphi_to_pos, cosphi_from_pos

try:
    import pandaplan.core.pplog as logging
except ImportError:
    import logging

logger = logging.getLogger(__name__)


# -------------------------------------------------------------------------------------------------
# Basic Classes
# -------------------------------------------------------------------------------------------------


class BaseModel:
    @classmethod
    def name(cls):
        return str(cls).split(".")[-1][:-2]


class QVCurve:
    """ Simple Q(V) controller. The characteristic curve is defined by 'vm_points_pu' and
    'q_points_pu' (relative to sn_mva).

                                   - Q(Vm)/sn (underexcited)
                                   ^
                                   |
                                   |
                                   |               _______
                                   |              /
                                   |             /
                                   |            /
                                   |           /
                                   |          /
             v[0] v[1]   v[2]      |         /
       --------+----+-----+--------+--------+-----+------+------>
                         /         |      v[3]   v[4]   v[5]    Vm
                        /          |
                       /           |
                      /            |
                     /             |
                ____/              |
                                   + Q(Vm)/sn (overexcited)
    """
    def __init__(self, vm_points_pu, q_points_pu):
        self.vm_points_pu = vm_points_pu
        self.q_points_pu = q_points_pu

    def step(self, vm_pu):
        return np.interp(vm_pu, self.vm_points_pu, self.q_points_pu)


class CosphiVCurve:
    """ Simple Q(V) controller. The characteristic curve is
    defined by 'vm_points_pu' and 'cosphi_points' (pos. values -> overexcited, voltage increasing,
    neg. values -> underexcited, voltage decreasing).

                                   - cosphi(Vm) (underexcited)
                                   ^
                                   |
                                   |
                                   |               _______
                                   |              /
                                   |             /
                                   |            /
                                   |           /
                                   |          /
             v[0] v[1]   v[2]      |         /
       --------+----+-----+--------+--------+-----+------+------>
                         /         |      v[3]   v[4]   v[5]    Vm
                        /          |
                       /           |
                      /            |
                     /             |
                ____/              |
                                   + cosphi(Vm) (overexcited)
    """
    def __init__(self, vm_points_pu, cosphi_points):
        self.vm_points_pu = vm_points_pu
        self.cosphi_points = cosphi_points
        self.cosphi_pos = cosphi_to_pos(self.cosphi_points)

    def step(self, vm_pu, p_pu):
        cosphi = cosphi_from_pos(np.interp(vm_pu, self.vm_points_pu, self.cosphi_pos))
        return np.tan(np.arccos(cosphi)) * p_pu


class CosphiPCurve:
    """ CosphiPCurve is a Q(P) controller, more precisely a cosphi(P) controller. The characteristic
    curve is defined by 'p_points_pu' and 'cosphi_points' (pos. values -> overexcited, voltage
    increasing, neg. values -> underexcited, voltage decreasing).

    **Exemplary Characteristic curve**::

        cosphi(P)
          ^
          |
          |________
          |        .
          |         .
          |          .
          |           .
          |            . ___________
          |
          |
        0 +--+-----+----+-------------->
           p[0]    p[1]  p[2]
    INPUT:
        **p_points_pu** (iterable of floats) - active power values (relative to sn_mva) on the
        cosphi(P) curve.

        **cosphi_points** (iterable of floats) - cosphi values on the cosphi(P) curve. positive
        values lead to positive reactive power values (inductive/overexcited generator), negative
        values mean capactive/underexcited generator
    """
    def __init__(self, p_points_pu, cosphi_points):
        self.p_points_pu = np.array(p_points_pu)
        self.cosphi_points = np.array(cosphi_points)
        if any(self.cosphi_points > 1):
            raise ValueError("cosphi cannot be higher than 1")
        if any(self.cosphi_points < -1):
            raise ValueError("cosphi cannot be lower than -1")

        self.cosphi_pos = cosphi_to_pos(self.cosphi_points)

    def step(self, p_pu):
        cosphi = cosphi_from_pos(np.interp(p_pu, self.p_points_pu, self.cosphi_pos))
        return np.tan(np.arccos(cosphi)) * p_pu


if __name__ == "__main__":
    pass
